'-------------------------------- Loop05 ---------------------------------'
'                                                                         '
' This program solves the time dependent heat conduction equation         '
' numerically using the Finite Volume Method.  The linear system that     '
' results is solved using the Gauss Seidel Iteration.  The time           '
' dependence is computed using the fully implicit (and most compute       '
' costly) method.  The program compares a procedure using rolled-up       '
' loops to one using unrolled loops.                                      '
'                                                                         '
' NOTE:  This program as written differs from the Windows version (and    '
'        in the book) in the end time.  The book and Windows versions     '
'        use an end time of 1200 seconds.  This version, written for      '
'        DOS compiler, uses an end time of 600 seconds.                   '
'                                                                         '
'-------------------------------------------------------------------------'

'global declarations
'longs
DIM j           AS LONG                 'j loop counter
DIM k           AS LONG                 'k loop counter
DIM counter     AS LONG                 'generic counter
DIM NumIts      AS LONG                 'Number of iterations

'singles
DIM tStep       AS SINGLE               'time step
DIM tStop       AS SINGLE               'ending time
DIM OptTol      AS SINGLE               'convergence tolerance
DIM A(5, 5)     AS SINGLE               'coefficient matrix
DIM b(5)        AS SINGLE               'right side 'vector'
DIM T0(5)       AS SINGLE               'node temps at start of time step
DIM T(5)        AS SINGLE               'node temps
DIM delta.T     AS SINGLE               'max change in node temp at end of iteration
DIM max.delta.T AS SINGLE               'maximum node temp change during iteration
DIM T.Old(5)    AS SINGLE               'node temps, last iteration
DIM Tdiff       AS SINGLE               'change in node temp
DIM Sum1        AS SINGLE               'part of sol'n using solved nodes
DIM Sum2        AS SINGLE               'part of sol'n using unsolved nodes
DIM Tcounter    AS SINGLE               'temperature counter
DIM starttime   AS SINGLE               'ticks at start of solution
DIM stoptime    AS SINGLE               'tick when finished
DIM AP0         AS SINGLE               'node factor


'------------------------------ Initializations --------------------------

'time step parameters
tStep = .001
tStop = 600

'solver convergence tolerance
OptTol = .00001

'initial conditions
T0(1) = 200
T0(2) = 200
T0(3) = 200
T0(4) = 200
T0(5) = 200

T(1) = T0(1)
T(2) = T0(2)
T(3) = T0(3)
T(4) = T0(4)
T(5) = T0(5)

'linear system coefficient matrix, A
A(1, 1) = 400025
A(1, 2) = -25
A(1, 3) = 0
A(1, 4) = 0
A(1, 5) = 0

A(2, 1) = -25
A(2, 2) = 400050
A(2, 3) = -25
A(2, 4) = 0
A(2, 5) = 0

A(3, 1) = 0
A(3, 2) = -25
A(3, 3) = 400050
A(3, 4) = -25
A(3, 5) = 0

A(4, 1) = 0
A(4, 2) = 0
A(4, 3) = -25
A(4, 4) = 400050
A(4, 5) = -25

A(5, 1) = 0
A(5, 2) = 0
A(5, 3) = 0
A(5, 4) = -25
A(5, 5) = 400075

AP0 = 10000000# * (.004 / tStep) * .01

'-------------------------- End Initializations -----------------------------

'-------------------------- Solver Using Rolled Loops -----------------------

'clear the screen
CLS

'display message
PRINT "Solving system from 2 to 600 seconds using rolled loops ..."
PRINT


'start the timer
starttime = TIMER

'begin solving
FOR Tcounter = 2 TO tStop STEP tStep

    'set 'right side' of matrix equation
    FOR counter = 1 TO 5
    
	b(counter) = AP0 * T0(counter)
	
    NEXT counter
    
    'solve the system for T(j)
    'using Gauss-Seidel procedure
    NumIts = 1
    delta.T = 1
    max.delta.T = 0
   
    DO WHILE delta.T >= OptTol AND NumIts <= 1000
	
	FOR counter = 1 TO 5
    
	    T.Old(counter) = T(counter)
       
	NEXT counter
    
	FOR j = 1 TO 5

	    Sum1 = 0
	   
	    FOR k = 1 TO j - 1
	
		Sum1 = A(j, k) * T(k) + Sum1
	    
	    NEXT k
	
	    Sum2 = 0
	
	    FOR k = j + 1 TO 5
	
		Sum2 = Sum2 + (A(j, k) * T.Old(k))
	    
	    NEXT k
	
	    T(j) = Sum1 + Sum2 - b(j)
	
	    T(j) = T(j) / -A(j, j)
	    
	    Tdiff = ABS(T(j) - T.Old(j))

	    IF Tdiff > max.delta.T THEN
 
	       max.delta.T = Tdiff

	    END IF
	
	NEXT j
    
	NumIts = NumIts + 1
	delta.T = max.delta.T
	max.delta.T = 0

    LOOP
    
    FOR counter = 1 TO 5
    
	T0(counter) = T(counter)
	
    NEXT counter

NEXT Tcounter

'stop the timer
stoptime = TIMER

'Display the result
PRINT "Via Rolled Loops, final values of Node Temperatures were:"
PRINT "T(1)= "; T(1); "  T(2)= "; T(2)
PRINT "T(3)= "; T(3); "  T(4)= "; T(4)
PRINT "T(5)= "; T(5)
PRINT "Solved in "; stoptime - starttime; " seconds"
PRINT

'---------------------- End Solver Using Rolled Loops -----------------------

'------------------------ Solver Using Unrolled Loops -----------------------

'reset initial conditions initial
T0(1) = 200
T0(2) = 200
T0(3) = 200
T0(4) = 200
T0(5) = 200

T(1) = T0(1)
T(2) = T0(2)
T(3) = T0(3)
T(4) = T0(4)
T(5) = T0(5)
    
'display message
PRINT "Solving system from 2 to 600 seconds using unrolled loops ..."
PRINT
    
'start the timer
starttime = TIMER

FOR Tcounter = 2 TO tStop STEP tStep
    
    'set 'right side' of matrix equation
    b(1) = AP0 * T0(1)
    b(2) = AP0 * T0(2)
    b(3) = AP0 * T0(3)
    b(4) = AP0 * T0(4)
    b(5) = AP0 * T0(5)
    
    'solve the system for T(j)
    'using Gauss-Seidel procedure
    NumIts = 1
    delta.T = 1
    max.delta.T = 0

    DO WHILE delta.T >= OptTol AND NumIts <= 1000
    
	T.Old(1) = T(1)
	T.Old(2) = T(2)
	T.Old(3) = T(3)
	T.Old(4) = T(4)
	T.Old(5) = T(5)
  
	'this begins the unrolled Gauss-Seidel procedure
	'j=1
	'k=2-5
	Sum2 = A(1, 2) * T.Old(2) + A(1, 3) * T.Old(3) + A(1, 4) * T.Old(4) + A(1, 5) * T.Old(5)
	T(1) = Sum2 - b(1)
	T(1) = T(1) / -A(1, 1)
	
	max.delta.T = ABS(T(1) - T.Old(1))

	'j=2
	'k=1
	Sum1 = A(2, 1) * T(1)
	
	'k=3-5
	Sum2 = A(2, 3) * T.Old(3) + A(2, 4) * T.Old(4) + A(2, 5) * T.Old(5)

	T(2) = Sum1 + Sum2 - b(2)
	T(2) = T(2) / -A(2, 2)
	
	Tdiff = ABS(T(2) - T.Old(2))
	
	IF Tdiff > max.delta.T THEN
	   
	    max.delat.T = Tdiff

	END IF
	
	'j=3
	'k=1-2
	Sum1 = A(3, 1) * T(1) + A(3, 2) * T(2)
	
	'k=4-5
	Sum2 = A(3, 4) * T.Old(4) + A(3, 5) * T.Old(5)

	T(3) = Sum1 + Sum2 - b(3)
	T(3) = T(3) / -A(3, 3)
	
	Tdiff = ABS(T(3) - T.Old(3))

	IF Tdiff > max.delta.T THEN

	    max.delta.T = Tdiff

	END IF
	
	'j=4
	'k=1-3
	Sum1 = A(4, 1) * T(1) + A(4, 2) * T(2) + A(4, 3) * T(3)
	
	'k=5
	Sum2 = A(4, 5) * T.Old(5)

	T(4) = Sum1 + Sum2 - b(4)
	T(4) = T(4) / -A(4, 4)
	
	Tdiff = ABS(T(4) - T.Old(4))

	IF Tdiff > max.delta.T THEN

	    max.delta.T = Tdiff

	END IF
	
	'j=5
	'k=1-4
	Sum1 = A(5, 1) * T(1) + A(5, 2) * T(2) + A(5, 3) * T(3) + A(5, 4) * T(4)
	
	T(5) = Sum1 - b(5)
	T(5) = T(5) / -A(5, 5)
	
	Tdiff = ABS(T(5) - T.Old(5))

	IF Tdiff > max.delta.T THEN

	    max.delta.T = Tdiff

	END IF
    
	NumIts = NumIts + 1
	delta.T = max.delta.T
	max.delta.T = 0

    LOOP
    
    T0(1) = T(1)
    T0(2) = T(2)
    T0(3) = T(3)
    T0(4) = T(4)
    T0(5) = T(5)

NEXT Tcounter
    
'stop the timer
stoptime = TIMER
    
'display results
PRINT "Via Unrolled Loops, Node Temperatures:"
PRINT "T(1)= "; T(1); "  T(2)= "; T(2)
PRINT "T(3)= "; T(3); "  T(4)= "; T(4)
PRINT "T(5)= "; T(5)
PRINT "Solved in "; stoptime - starttime; " seconds"

'-------------------- End Solver Using Unrolled Loops -----------------------

' Keep display active until user presses a key
' this is needed when running in the dos subsystem of windows, else
' windows closes window immediately

' display a message
PRINT : PRINT "Press any key to end..."

'check keylatch; if empty, simply loop back
LoopHere:
In$ = INKEY$
IF In$ = "" THEN GOTO LoopHere

'end when keylatch contains any key
END

