// Loop04.c: implementation file
//
// This program compares the execution times for a repetitive Simpson's
// rule integration with the summation loop fully rolled up or partially
// unrolled.  The technique of partial unrolling is useful when full
// unrolling is not feasible.
//
// Recall that the integral, I, of a function is divided into 2n steps
// (Simpson's Rule requires an even number of steps) can be computed
// by
//
//				I = h/3 * (s0 + 2s1 + 4s0)
//
// where h = (b-a)/2n, s0 = f(0) + f(n), s1 = f(1) + f(3) + f(5) + ...
// f(2n-2) and s2 = f(2) + f(4) + f(6) + ... + f(2n - 1)
//
// The rolled-up Simpson's Rule is performed in fnA().
//
// The Simpson's Rule unrolled four terms is performed in fnB().
//
// The Simpson's Rule unrolled five terms is performed in fnC().
//
// The Simpson's Rule unrolled eight terms is performed in fnD().

#include <sys/time.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

//defines
#define NumIts 50000		//number of iterations

//functions
void fnA();			
void fnB();
void fnC();
void fnD();

// global declarations
float f[10000];				//function f(x)
 			

//Rolled-Up Simpson's Rule Integration of f(x)
void fnA()
{

	//local declarations
	//integers
	int SimpCounter;		//simpson's rule summation counter
	int ItCounter;			//iteration counter
	int Counter;			//function assignment counter

	//floats
	float h_3;				//simpson's rule parameters
	float S0;
	float S1;
	float S2;
	float Integral;			//result

	//doubles
	double argX;			//x for assignment
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//assign the function f(x) = x^5
	for(Counter=1; Counter <= 10000; ++Counter){

		argX = (double)Counter * 0.0001;
		f[Counter] = (float)pow(argX, 5);

	}

	//display entry message
	printf("\n Rolled-Up Simpson's Rule Integration ...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);

	for(ItCounter=1; ItCounter <= NumIts; ++ItCounter){

		//h/3, where h = (b-a)/n
		h_3 = (float)0.999 / 30000;

		//sum the first and last terms
		S0 = f[0] + f[9999];

		//zero values for running summations
		S1 = 0;
		S2 = 0;

		//sum terms except first and last
		for(SimpCounter = 1; SimpCounter <= 9997; SimpCounter=SimpCounter + 2){

			S1 += f[SimpCounter];
			S2 += f[SimpCounter + 1];

		}

		// 'odd' terms * 2
		S1 *= 2;

		//'even' terms * 4
		S2 *= 4;

		//total summation
		Integral = S0 + S1 + S2;

		//multiply by h/3
		Integral *= h_3;

	}

	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	printf("%s%g%s", "Integral = ", Integral, "\n");

}//end fnA()

// Simpson's Rule Integration Unrolled by four explicit terms 
void fnB()
{

	//local declarations
	//integers
	int SimpCounter;		//simpson's rule summation counter
	int ItCounter;			//iteration counter
	int Counter;			//function assignment counter

	//floats
	float h_3;				//simpson's rule parameters
	float S0;
	float S1;
	float S2;
	float Integral;			//result

	//doubles
	double argX;			//x for assignment
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//assign the function f(x) = x^5
	for(Counter=1; Counter <= 10000; ++Counter){

		argX = (double)Counter * 0.0001;
		f[Counter] = (float)pow(argX, 5);

	}

	//display entry message
	printf("\n Simpson's Rule Integration Unrolled Four Terms...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);

	for(ItCounter=1; ItCounter <= NumIts; ++ItCounter){

		//h/3, where h = (b-a)/n
		h_3 = (float)0.999 / 30000;

		//sum the first and last terms
		S0 = f[0] + f[9999];

		//zero values for running summations
		S1 = 0;
		S2 = 0;

		//unrolled summation
		for(SimpCounter = 1; SimpCounter <= 9985; SimpCounter += 8){

			S1 += f[SimpCounter] + f[SimpCounter + 2] + f[SimpCounter + 4] + f[SimpCounter + 6];
			S2 += f[SimpCounter + 1] + f[SimpCounter + 3] + f[SimpCounter + 5] + f[SimpCounter + 7];

		}

		//clean up extra terms
		S1 += f[9993] + f[9995] + f[9997];
		S2 += f[9994] + f[9996] + f[9998];

		// 'odd' terms * 2
		S1 *= 2;

		//'even' terms * 4
		S2 *= 4;

		//total summation
		Integral = S0 + S1 + S2;

		//multiply by h/3
		Integral *= h_3;

	}

	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	printf("%s%g%s", "Integral = ", Integral, "\n");
	
}//end fnB()

// Simpson's rule integration unrolled by five explicit terms
void fnC()
{

	//local declarations
	//integers
	int SimpCounter;		//simpson's rule summation counter
	int ItCounter;			//iteration counter
	int Counter;			//function assignment counter

	//floats
	float h_3;				//simpson's rule parameters
	float S0;
	float S1;
	float S2;
	float Integral;			//result

	//doubles
	double argX;			//x for assignment
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//assign the function f(x) = x^5
	for(Counter=1; Counter <= 10000; ++Counter){

		argX = (double)Counter * 0.0001;
		f[Counter] = (float)pow(argX, 5);

	}

	//display entry message
	printf("\n Simpson's Rule Integration Unrolled Five terms...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);

	for(ItCounter=1; ItCounter <= NumIts; ++ItCounter){

		//h/3, where h = (b-a)/n
		h_3 = (float)0.999 / 30000;

		//sum the first and last terms
		S0 = f[0] + f[9999];

		//zero values for running summations
		S1 = 0;
		S2 = 0;

		//unrolled summation
		for(SimpCounter = 1; SimpCounter <= 9981; SimpCounter += 10){
		
			S1 += f[SimpCounter] + f[SimpCounter + 2] + f[SimpCounter + 4] + f[SimpCounter + 6] + f[SimpCounter + 8];
			S2 += f[SimpCounter + 1] + f[SimpCounter + 3] + f[SimpCounter + 5] + f[SimpCounter + 7] + f[SimpCounter + 9];

		}

		//clean up extra terms
		S1 += f[9991] + f[9993] + f[9995] + f[9997];
		S2 += f[9992] + f[9994] + f[9996] + f[9998];

		// 'odd' terms * 2
		S1 *= 2;

		//'even' terms * 4
		S2 *= 4;

		//total summation
		Integral = S0 + S1 + S2;

		//multiply by h/3
		Integral *= h_3;

	}

	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	printf("%s%g%s", "Integral = ", Integral, "\n");
	
}//end fnC()

// Simpson's Rule Integration unrolled eight explicit terms 
void fnD()
{

	//local declarations
	//integers
	int SimpCounter;		//simpson's rule summation counter
	int ItCounter;			//iteration counter
	int Counter;			//function assignment counter

	//floats
	float h_3;				//simpson's rule parameters
	float S0;
	float S1;
	float S2;
	float Integral;			//result

	//doubles
	double argX;			//x for assignment
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//assign the function f(x) = x^5
	for(Counter=1; Counter <= 10000; ++Counter){

		argX = (double)Counter * 0.0001;
		f[Counter] = (float)pow(argX, 5);

	}

	//display entry message
	printf("\n Simpson's Rule Integration Unrolled Eight terms...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);

	for(ItCounter=1; ItCounter <= NumIts; ++ItCounter){

		//h/3, where h = (b-a)/n
		h_3 = (float)0.999 / 30000;

		//sum the first and last terms
		S0 = f[0] + f[9999];

		//zero values for running summations
		S1 = 0;
		S2 = 0;

		//unrolled summation
		for(SimpCounter = 1; SimpCounter <= 9969; SimpCounter += 16){
		
			S1 += f[SimpCounter] + f[SimpCounter + 2] + f[SimpCounter + 4] + f[SimpCounter + 6] + f[SimpCounter + 8]
					+ f[SimpCounter + 10] + f[SimpCounter + 12] + f[SimpCounter + 14];
			S2 += f[SimpCounter + 1] + f[SimpCounter + 3] + f[SimpCounter + 5] + f[SimpCounter + 7] + f[SimpCounter + 9]
					+ f[SimpCounter + 11] + f[SimpCounter + 13] + f[SimpCounter + 15];

		}

		//clean up extra terms
		S1 += f[9985] + f[9987] + f[9989] + f[9991] + f[9993] + f[9995] + f[9997];
		S2 += f[9986] + f[9988] + f[9990] + f[9992] + f[9994] + f[9996] + f[9998];

		// 'odd' terms * 2
		S1 *= 2;

		//'even' terms * 4
		S2 *= 4;

		//total summation
		Integral = S0 + S1 + S2;

		//multiply by h/3
		Integral *= h_3;

	}

	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	printf("%s%g%s", "Integral = ", Integral, "\n");
	
}//end fnD()

//entry function
int main()
{
	//local declarations
	//integers
	int bTERM;		//menu flag

	//character strings
	char inchar;		//input character
	

	//initialize bTERM to NOT TERMINATE
	bTERM = 0;
	
	//clear terminal
	system("clear");
	
	do {
		
		//present menu to user
		printf("Enter the Number of the routine to run, and press <Enter> \n");
		printf("1\tRolled-Up Simpson's Rule Integration\n");
		printf("2\tSimpson's Rule Integration unrolled four explicit terms\n");
		printf("3\tSimpson's Rule Integration unrolled five explicit terms\n");
		printf("4\tSimpson's Rule Integration unrolled eight explicit terms\n");
		printf("q\tEnd this program\n");
	
		//get user input
		scanf("%s", &inchar);

		//call the user selected function
		switch	(inchar){

			case '1': fnA();
				break;
	
			case '2': fnB();
				break;

			case '3': fnC();
				break;

			case '4': fnD();
				break;
		
			//select the terminate function	
			case 'q': printf("Ending This Program\n");
				
				//set flag to terminate
				bTERM = 1;
				break;

		}
		
	} while(!bTERM);

	return 0;

}
