// Loop05.c : implementation file
//
// This program solves the time dependent heat equation numerically
// using the Finite Volume Method.  Gauss-Seidel iteration is used to
// solve the coupled equations.  The time dependence is modeled using
// the fully implicit (and more computationally costly) method.
//
// The rolled version of the solution is in fnA()
// The unrolled version is fnA()

#include <sys/time.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

//defines
#define tStep	0.001		//time step, in seconds
#define tStop	1200		//ending time, in seconds
#define OptTol	0.00001		//tolerance for convergence test


//This routine uses rolled up loops for the Gauss Seidel Solver
void fnA()
{

	//local declarations
	//integers
	int counter;			//handy counter for small loops
	int i;					//row counter in Gauss-Seidel solver
	int k;					//column counter in Gauss-Seidel solver
	int NumIts;				//number of iterations in Gauss-Seidel solver
							//this is used to prevent infinite loop in case
							//of poor or non-convergence in the solver

	//floats
	float delta_T;			//max change in Temperature; used to test for convergence
	float max_delta_T;		//used to determine the delta_T
	float sum1;				//a temporary summation holder
	float sum2;				//a temporary summation holder
	float T_old[5];			//the set of temperatures at the last iteration
	float tCounter;			//time counter
	float Tdiff;			//individual Temperature difference; used
							//to compute max_delta_T

	float a[5][5];			//the coefficient matrix in the linear system
	float b[5];				//'right hand side' of linear system
	float T0[5];			//node temperatures at start of given time step
							//these are updated at the end of each time step
	float T[5];				//node temperatures at five spatial location
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end

	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//display entry message
	printf("\nNumerical Heat Conduction using rolled loops...\n");

	//initialize the linear system using data from Ch. 7
	//initial node temperatures
	for (counter=0; counter <= 4; ++counter){

		T0[counter] = 200;
		T[counter] = 200;

	}

	//coefficient matrix
	a[0][0] = 400025;
	a[0][1] = -25;
	a[0][2] = 0;
	a[0][3] = 0;
	a[0][4] = 0;

	a[1][0] = -25;
	a[1][1] = 400050;
	a[1][2] = -25;
	a[1][3] = 0;
	a[1][4] = 0;

    a[2][0] = 0;
	a[2][1] = -25;
	a[2][2] = 400050;
	a[2][3] = -25;
	a[2][4] = 0;

	a[3][0] = 0;
	a[3][1] = 0;
	a[3][2] = -25;
	a[3][3] = 400050;
	a[3][4] = -25;

	a[4][0] = 0;
	a[4][1] = 0;
	a[4][2] = 0;
	a[4][3] = -25;
	a[4][4] = 400075;
	
	
	
	//get start time
	gettimeofday(&_tstart, &tz);
	
	for (tCounter = (float)0.001; tCounter <= (float)tStop; tCounter += (float)tStep){

		//set right side of the system
		for(counter = 0; counter <=4; ++counter){

			b[counter] = 400000 * T0[counter];

		}

		//solve the system for T(j) using Gauss-Seidel procedure

		NumIts = 1;
		delta_T = 1;
		max_delta_T = 0;

		do{

			//save the old T(j)'s
			for(counter = 0; counter <= 4; ++counter){

				T_old[counter] = T[counter];

			}

			for(i=0; i<= 4; ++i){

				sum1 = 0;

				for(k=0; k <= i-1; ++k){

					sum1 += a[i][k]*T[k];

				}

				sum2 = 0;

				for(k = i+1; k<=4; ++k){

					sum2 += a[i][k]*T_old[k];

				}

				T[i] = sum1 + sum2 - b[i];
				T[i] /= -a[i][i];

				Tdiff = (float)fabs(T[i] - T_old[i]);

				if(Tdiff > max_delta_T){

					max_delta_T = Tdiff;

				}

			}

			++NumIts;
			delta_T = max_delta_T;
			max_delta_T = 0;

		}while ((delta_T >= OptTol) & (NumIts <= 1000));

		//update the 'initial' temperatures for the beginning
		//of the next time step

		for(counter = 0; counter <=4; ++counter){

			T0[counter] = T[counter];

		}

	}
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");

}//end fnA()

///this routine uses unrolled loops for the Gauss-Seidel solver 
void fnB()
{

	//local declarations
	//integers
	int counter;			//handy counter for small loops
	int NumIts;				//number of iterations in Gauss-Seidel solver
							//this is used to prevent infinite loop in case
							//of poor or non-convergence in the solver

	//floats
	float delta_T;			//max change in Temperature; used to test for convergence
	float max_delta_T;		//used to determine the delta_T
	float sum1;				//a temporary summation holder
	float sum2;				//a temporary summation holder
	float T_old[5];			//the set of temperatures at the last iteration
	float tCounter;			//time counter
	float Tdiff;			//individual Temperature difference; used
							//to compute max_delta_T

	float a[5][5];			//the coefficient matrix in the linear system
	float b[5];				//'right hand side' of linear system
	float T0[5];			//node temperatures at start of given time step
							//these are updated at the end of each time step
	float T[5];				//node temperatures at five spatial location
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end

	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//display entry message
	printf("\nNumerical Heat Conduction using unrolled loops...\n");

	//initialize the linear system using data from Ch. 7
	//initial node temperatures
	for (counter=0; counter <= 4; ++counter){

		T0[counter] = 200;
		T[counter] = 200;

	}

	//coefficient matrix
	a[0][0] = 400025;
	a[0][1] = -25;
	a[0][2] = 0;
	a[0][3] = 0;
	a[0][4] = 0;

	a[1][0] = -25;
	a[1][1] = 400050;
	a[1][2] = -25;
	a[1][3] = 0;
	a[1][4] = 0;

    a[2][0] = 0;
	a[2][1] = -25;
	a[2][2] = 400050;
	a[2][3] = -25;
	a[2][4] = 0;

	a[3][0] = 0;
	a[3][1] = 0;
	a[3][2] = -25;
	a[3][3] = 400050;
	a[3][4] = -25;

	a[4][0] = 0;
	a[4][1] = 0;
	a[4][2] = 0;
	a[4][3] = -25;
	a[4][4] = 400075;
	
	
	
	//get start time
	gettimeofday(&_tstart, &tz);
	
	for (tCounter = (float)0.001; tCounter <= (float)tStop; tCounter += (float)tStep){

		//set right side of the system
		for(counter = 0; counter <=4; ++counter){

			b[counter] = 400000 * T0[counter];

		}

		//solve the system for T(j) using Gauss-Seidel procedure

		NumIts = 1;
		delta_T = 1;
		max_delta_T = 0;

		do{

			//save the old T(j)'s

			T_old[0] = T[0];
			T_old[1] = T[1];
			T_old[2] = T[2];
			T_old[3] = T[3];
			T_old[4] = T[4];

			//this begins the unrolled Gauss-Seidel procedure
			
			//i=0
			//k=1-4
			sum2 = a[0][1]*T_old[1] + a[0][2]*T_old[2] + a[0][3]*T_old[3] + a[0][4]*T_old[4];

			T[0] = sum2 - b[0];
			T[0] /= -a[0][0];

			max_delta_T = (float)fabs(T[0] - T_old[0]);

			//i=1
			//k=0

			//k=2-4
			sum2 = a[1][2]*T_old[2] + a[1][3]*T_old[3] + a[1][4]*T_old[4];
			
			T[1] = a[1][0]*T[0] + sum2 - b[1];
			T[1] /= -a[1][1];

			Tdiff = (float)fabs(T[1] - T_old[1]);
				
			if(Tdiff > max_delta_T){

				max_delta_T = Tdiff;

			}

			//i=2
			//k=0-1
			sum1 = a[2][0]*T[0] + a[2][1]*T[1];

			//k=3-4
			sum2 = a[2][3]*T_old[3] + a[2][4]*T_old[4];

			T[2] = sum1 + sum2 - b[2];
			T[2] /= -a[2][2];

			Tdiff = (float)fabs(T[2] - T_old[2]);
				
			if(Tdiff > max_delta_T){

				max_delta_T = Tdiff;

			}

			//i=3
			//k=0-2
			sum1 = a[3][0]*T[0] + a[3][1]*T[1] + a[3][2]*T[2];

			T[3] = sum1 + a[3][4]*T[4] - b[3];
			T[3] /= -a[3][3];

			Tdiff = (float)fabs(T[3] - T_old[3]);
				
			if(Tdiff > max_delta_T){

				max_delta_T = Tdiff;
			
			}
				
			//i=4
			//k=0-3
			sum1 = a[4][0]*T[0] + a[4][1]*T[1] + a[4][2]*T[2] + a[4][3]*T[3];

			T[4] = sum1 - b[4];
			T[4] /= -a[4][4];

			Tdiff = (float)fabs(T[4] - T_old[4]);
				
			if(Tdiff > max_delta_T){

				max_delta_T = Tdiff;

			}

			++NumIts;
			delta_T = max_delta_T;

		}while ((delta_T >= OptTol) & (NumIts <= 1000));

		//update the 'initial' temperatures for the beginning
		//of the next time step

		T0[0] = T[0];
		T0[1] = T[1];
		T0[2] = T[2];
		T0[3] = T[3];
		T0[4] = T[4];

	}
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	
}//end fnB()

//entry function
int main()
{
	//local declarations
	//integers
	int bTERM;		//menu flag

	//character strings
	char inchar;		//input character
	

	//initialize bTERM to NOT TERMINATE
	bTERM = 0;
	
	//clear terminal
	system("clear");
	
	do {
		
		//present menu to user
		printf("Enter the Number of the routine to run, and press <Enter> \n");
		printf("1\tNumerical Heat Conduction with Rolled Loops\n");
		printf("2\tNumerical Heat Conduction with Unrolled Loops\n");
		printf("q\tEnd this program\n");
	
		//get user input
		scanf("%s", &inchar);

		//call the user selected function
		switch	(inchar){

			case '1': fnA();
				break;
	
			case '2': fnB();
				break;

			//select the terminate function	
			case 'q': printf("Ending This Program\n");
				
				//set flag to terminate
				bTERM = 1;
				break;

		}
		
	} while(!bTERM);

	return 0;

}
