// Other02.c : implementation file
//
// This program demonstrates the elimination of nested loop code
// for an algorithm amenable to this optimization.
//
// In the Monte Carlo procedure, the total pair energy
// is not needed at each step; only the change in energy
// that results from moving a single particle is needed.
//
// The "Full Energy" using nested loops is computed in fnB().
// The energy using only "Single Particle Energy" is computed in fnC().
// Look Up Tables and other initializations are done in fnA().

#include <sys/time.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>

//defines
#define NUMITS 50000		//number of iterations

//global declarations

//integers
int Rval[354][354];         //Look-Up Table Entry for precomputed distances
int bINIT;		    //initializations have been done
              
//floats
float Sigma;                //LJ Sigma parameter
float Depth;                //LJ Well depth parameter

//doubles
double Energy[354];         //Look-Up Table Entry for the 6 LJ Energies as
                            //as function of distance; distance mesh is 10pm
double TotalEnergy;         //Total Energy

//structures
struct particle {           //Structure to hold particle data; just a bit neater
                            //to store this way, and sets stage for OOP if desired
    float x;                //x coordinate
    float y;                //y coordinate
    
} Particles[20];

//functions
void fnA();			
void fnB();
void fnC();			

double calc_ljenergy(double R);		// the called function

//perform iterative function calling
void fnA()
{

    //local declarations
    //integers
    int PCounter;    //particle counter
    int rCountx;     //x coordinate counter
    int rCounty;     //y coordinate counter
    
    //floats
    float tempx;     //cache variable for x
    float tempy;     //cache variable for y

    //doubles
    double R;         //Cache variable for distance
    double R6;        //Distance^6
    double R12;       //Distance^12
    double Sigma6;    //LJ Sigma^6
    double Sigma12;   //LJ Sigma^12
    

    for(rCountx = 0; rCountx < 354; ++rCountx){

	//distance mesh (or step) is 10 pm
        R = rCountx * 10;
        
        //compute distance^6
        R6 = R * R * R * R * R * R;
        
        //compute distance^12
        R12 = R6 * R6;
        
        //Compute Sigma^6
        Sigma6 = Sigma * Sigma * Sigma * Sigma * Sigma * Sigma;
        
        //compute Sigma^12
        Sigma12 = Sigma6 * Sigma6;
        
        //Compute LJ Energy
        Energy[rCountx] = Sigma12 / R12 - Sigma6 / R6;
        Energy[rCountx] = Energy[rCountx] * Depth;
        
	//distance mesh is 10 pm
        tempx = (float)rCountx * 10;
        
        //square "delta-x"
        tempx = tempx * tempx;
        
        for(rCounty=0; rCounty < 354; ++rCounty){
        
            //distance mesh is 10 pm
            tempy = (float)rCounty * 10;
            
            //square "delta-y"
            tempy = tempy * tempy;

            //store the distance
            Rval[rCountx][rCounty] = (int)(sqrt(tempx + tempy) * 0.1);
            
        }
    
    }
        
    //initialize random particle positions
    for(PCounter =0; PCounter < 20; ++PCounter){

        Particles[PCounter].x = (float)(rand()/RAND_MAX);
        Particles[PCounter].y = (float)(rand()/RAND_MAX);
        
    }

    bINIT = 1;
	
}//end fnA()

// full energy routine
void fnB()
{

	//local declarations
	//integers
	int counter;		//iteration counter
	int Ipart;          	//Particle I counter
    	int Jpart;          	//Particle J counter
    	int deltaX;         	//Difference between x of I and J particles
   	int deltaY;        	//Difference between y of I and J particles
    	int R;              	//R from Rval lookup table

	//floats
	float sXi;		//cache variable for x of particle I
	float sYi;		//cache variable for y of particle I
	float sXj;		//cache variable for x of particle J
	float sYj;		//cache variable for y of particle J
	
	//doubles
	double ThisEnergy;  //running variable for energy
    
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;
		
	//display entry message
	printf("\nIterative Full Energy Calculation...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);
	
		for(counter = 1; counter <= NUMITS; ++counter){
                
	    	  //initialize to zero
		  ThisEnergy = 0;

		  //compute energy in nested loop
		  for(Ipart = 0; Ipart < 19; ++Ipart){
    
			//cache x,y values
			sXi = Particles[Ipart].x;
			sYi = Particles[Ipart].y;
        
			for(Jpart = Ipart + 1; Jpart < 20; ++Jpart){
        
				//cache x,y values
				sXj = Particles[Jpart].x;
				sYj = Particles[Jpart].y;
            
				//compute x difference to 10 pm
				deltaX = (int)((fabs(sXi - sXj)) / 10);
            
				//compute y difference, to 10 pm
				deltaY = (int)((fabs(sYi - sYj)) / 10);
                
				//lookup distance rather than compute
				R = Rval[deltaX][deltaY];
        
				//lookup energy
				ThisEnergy = ThisEnergy + Energy[R];
        
			}
    
		}

	}
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	
}//end fnB()

void fnC(){

	//local declarations
	//integers
	int counter;		//iteration counter
	int I;	            //Particle I counter
 
        int deltaX;         //Difference between x of I and J particles
        int deltaY;         //Difference between y of I and J particles
        int R;              //R from Rval lookup table
	int MovedIndex;	    //index of moved particle

	//floats
	float sCurrentX;	//cache variable for x of particle I
	float sCurrentY;	//cache variable for y of particle I
	float sMovedX;		//cache for moved x
	float sMovedY;		//cache for moved y	
	//doubles
	double ThisEnergy;  //running variable for energy
    
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;
		
	//display entry message
	printf("\nIterative Full Energy Calculation...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);
	
	for(counter = 1; counter <= NUMITS; ++counter){
        
        	//select a particle
		MovedIndex = (int)(rand()/RAND_MAX);

		//cache coordinates of selected particle
		sMovedX = Particles[MovedIndex].x;
		sMovedY = Particles[MovedIndex].y;

		//initialize to zero
		ThisEnergy = 0;

		//compute energy in nested loop
		for(I = 0; I < 20; ++I){
    
			//cache x,y values
			sCurrentX = Particles[I].x;
			sCurrentY = Particles[I].y;
        
			if(I != MovedIndex){
        
				//compute x difference to 10 pm
				deltaX = (int)((fabs(sCurrentX - sMovedX)) / 10);
            
				//compute y difference, to 10 pm
				deltaY = (int)((fabs(sCurrentY - sMovedY)) / 10);
                
				//lookup distance rather than compute
				R = Rval[deltaX][deltaY];
        
				//lookup energy
				ThisEnergy = ThisEnergy + Energy[R];
        
			}
    
		}

	}

	
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");

}//end fnC()

//entry function
int main()
{
	//local declarations
	//integers
	int bTERM;		//menu flag

	//character strings
	char inchar;		//input character
	

	//initialize bINIT to NOT INITIATED
	bINIT = 0;

	//initialize bTERM to NOT TERMINATE
	bTERM = 0;
	
	//clear terminal
	system("clear");
	
	do {
		
		//present menu to user
		printf("Enter the Number of the routine to run, and press <Enter> \n");
		printf("1\tInitializations\n");

		if(bINIT){
		
			printf("2\tCompute Full Energy\n");
			printf("3\tCompute Only Delta Energy\n");
		
		}
		
		printf("q\tEnd this program\n");
	
		//get user input
		scanf("%s", &inchar);
		
		if((bINIT==0)&&((inchar=='2')||(inchar=='3'))){
		   
		   inchar=0;
		   
		}

		//call the user selected function
		switch	(inchar){

			case '1': fnA();
				break;
	
			case '2': fnB();
				break;
				
			case '3': fnC();
				break;
		
			//select the terminate function	
			case 'q': printf("Ending This Program\n");
				
				//set flag to terminate
				bTERM = 1;
				break;

		}
		
	} while(!bTERM);

	return 0;

} // End Main

double calc_ljenergy (double R)
{

	double R1;		//these would be used internally in the 
	double R2;		//optimized lj energy computation

	R1 = 0;
	R2 = 0;

	return(0);

}
