// Par01.c : implementation file
//

// This program compares the dot product using straight x87 floating
// point code to use of 3dNow! and SSE SIMD code.  If the run-time processor
// does not support 3dNow! or SSE, the SIMD routines simply exit with a 
// message (CPUid MUST be supported, however...I did not test for
// that support before making the calls to CPUid).
//
// The x87 routine is in fnA()
// The 3dNow! routine is in fnB()
// The SSE routine is in fnC()


#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

//defines
#define NumIts 100000000		//number of iterations

//functions
void fnA();			
void fnB();
void fnC();			


// perform dot product using x87 code.
// NOTE:  This is coded as inline assembly, but it is
// the output from the compiled unrolled loop C routine as
// in demo Data02.
void fnA()
{

	//general local declarations
	//integers
	int counter;			  // iteration counter

	//floats
	float a[4];		//vector a
	float b[4];		//vector b
	float c;		//dot product result
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//initialize the two vectors with some made up values
	a[0] = (float)23.45;
	a[1] = (float)14.65;
	a[2] = (float)33.87;
	a[3] = (float)0;

	b[0] = (float)36.99;
	b[1] = (float)21.32;
	b[2] = (float)17.38;
	b[3] = (float)0;
	
	
	//display entry message
	printf("\nComputing Unrolled Dot Product using only x87 ...\n");
	
	//get start time
	gettimeofday(&_tstart, &tz);
	
	for(counter=1; counter <= NumIts; ++counter){
    
		//c = a[0]*b[0] + a[1]*b[1] + a[2]*b[2] + a[3]*b[3];
		//compiles as:

		asm volatile(
			"flds    (%1)\n\t"
			"fmuls   (%2)\n\t"
			"flds   4(%1)\n\t"
			"fmuls  4(%2)\n\t"
			"faddp  %%st, %%st(1)\n\t"
			"flds   8(%1)\n\t"
			"fmuls  8(%2)\n\t"
			"faddp  %%st, %%st(1)\n\t"
			"flds	12(%1)\n\t"
			"fmuls	12(%2)\n\t"
			"faddp	%%st, %%st(1)\n\t"
			"fstps  %0"
			:"=m" (c)
			:"r" (a), "r"(b)
			
		);

    	}
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	printf("%s%g%s", "Dot product = ",c,"\n");
	
}//end fnA()

// perform dot product in ASSEMBLY using 3dNow! SIMD instructions
void fnB()
{

	//general local declarations
	//integers
	int counter;			//iteration counter
	int bAMDOK;			//cpu supports 3DNow! instructions


	//floats
	float a[4];			//vector a, 4-d
	float b[4];			//vector b, 4-d
	float c;			//dot product result
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//initialize AMD flag to zero
	bAMDOK = 0;

	//test for 3DNow! capability in cpu
	//this is adapted from code in AMD 3DNow! Tech. Manual
	asm volatile("pushl	%%eax\n\t"
		"pushl	%%edx\n\t"
		"xor	%%eax, %%eax\n\t"
		"CPUid\n\t"
		"test	%%eax, %%eax\n\t"
		"jz	No3dNow\n\t"	
		"movl	$0x80000000, %%eax\n\t"
		"CPUid\n\t"
		"cmp	$0x80000001, %%eax\n\t"
		"jl	No3dNow\n\t"
		"movl	$0x80000001, %%eax\n\t"
		"CPUid\n\t"
		"test	$0x80000000, %%edx\n\t"
		"jnz	Yes3dNow\n\t"
		"jmp	No3dNow\n"
"Yes3dNow:	\n\t"
		"movl	$1, %0\n\t"
		"jmp	EndAMDTest\n"
"No3dNow:	\n\t" 
		"movl	$0,%0\n"
"EndAMDTest:	\n\t"	 
		"popl	%%edx\n\t"
		"popl	%%eax"
		:"=g" (bAMDOK)
		:
		:"%eax","%ebx","ecx","%edx"
		
	);

	if(bAMDOK){

		//display entry message
		printf("\n4-d Dot Product using 3dNow! SIMD Code...\n");
	
		//initialize the two vectors with some made up values
		a[0] = (float)23.45;
		a[1] = (float)14.65;
		a[2] = (float)33.87;
		a[3] = (float)0;

		b[0] = (float)36.99;
		b[1] = (float)21.32;
		b[2] = (float)17.38;
		b[3] = (float)0;
		
		c=0;		
		
		//get start time
		gettimeofday(&_tstart, &tz);
	
		//set fp registers to SIMD mode
		asm ("femms\n\t");
	
		for(counter=1; counter <= NumIts; ++counter){
    
			//3DNow! dot product, using extended assembly lang.
			//We are doing 4-dimension vectors
			//here to avoid a movd instruction (doesn't matter
			//on athlon, but movd is slow on k6)
			//and we get the 4th d for 'free'
			asm volatile(
				"movq    (%1), %%mm0\n\t"
				"movq    (%2), %%mm1\n\t"
				"pfmul  %%mm1, %%mm0\n\t"
				"movq   8(%1), %%mm2\n\t"
				"movq   8(%2), %%mm1\n\t"
				"pfacc  %%mm0, %%mm0\n\t"
				"pfmul  %%mm2, %%mm1\n\t"
				"pfadd  %%mm1, %%mm0\n\t"
				"movd   %%mm0, %0"
				:"=m" (c)
				:"r" (a), "r" (b)	
					
			);
			
        
    		}
		

		//clear fp registers from SIMD mode
		asm ("femms\n\t");
	
		//stop the timer
		gettimeofday(&_tstop, &tz);

		//get seconds from the structures
		t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
		t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
		//compute the elapsed time
		ElapsedTime = (float)(t2 - t1);

		//display the result
		printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
		printf("%s%g%s","Dot Product = ",c,"\n\n");
	}
	
	else{
	
		printf("3dNow instructions not available on this CPU\n\n");
		
	}
	
	
}//end fnB()

// perform dot product in ASSEMBLY using SSE SIMD instructions
void fnC()
{

	//general local declarations
	//integers
	int counter;			//iteration counter
	int bSSEOK;			//cpu supports SSE instructions


	//floats
	float a[4];	//vector a, 4-d
	float b[4];	//vector b, 4-d
	float c[4];	//dot product result
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//initialize SSE flag to zero
	bSSEOK = 0;

	//test for SSE capability in cpu
	asm volatile(
	
		"pushl	%%eax\n\t"
		"pushl	%%edx\n\t"
		"xor	%%eax, %%eax\n\t"
		"CPUid\n\t"
		"test	%%eax, %%eax\n\t"
		"jz	NoSSE\n\t"	
		"movl	$0x00000001, %%eax\n\t"
		"CPUid\n\t"
		"test	$0x02000000, %%edx\n\t"
		"jnz	YesSSE\n\t"
		"jmp	NoSSE\n"
"YesSSE:	\n\t"
		"movl	$1, %0\n\t"
		"jmp	EndP3Test\n"
"NoSSE	:	\n\t" 
		"movl	$0,%0\n"
"EndP3Test:	\n\t"	 
		"popl	%%edx\n\t"
		"popl	%%eax"
		:"=g" (bSSEOK)
		:
		:"%eax","%ebx","%ecx","%edx"
		
	);

	if(bSSEOK){

		//display entry message
		printf("\n4-d Dot Product using SSE SIMD Code...\n");
	
		//initialize the two vectors with some made up values
		a[0] = (float)23.45;
		a[1] = (float)14.65;
		a[2] = (float)33.87;
		a[3] = (float)0;

		b[0] = (float)36.99;
		b[1] = (float)21.32;
		b[2] = (float)17.38;
		b[3] = (float)0;		
		
		//get start time
		gettimeofday(&_tstart, &tz);
	
		for(counter=1; counter <= NumIts; ++counter){
    
			//SSE dot product, using extended assembly lang.
			//technically, we are doing 4-dimension vectors
			//here to utilize the full 128 bits of SSE registers
			//we get the 4th d for free
			asm volatile(
				"movaps	(%1), %%xmm0\n\t"
				"movaps (%2), %%xmm1\n\t"
				"mulps  %%xmm1, %%xmm0\n\t"
				"movaps %%xmm0, %%xmm1\n\t"
				"shufps $0xee, %%xmm0, %%xmm1\n\t"
				"addps  %%xmm0, %%xmm1\n\t"
				"movaps %%xmm1, %%xmm0\n\t"
				"shufps $0xe5, %%xmm1, %%xmm1\n\t"
				"addps  %%xmm0, %%xmm1\n\t"
				"movss  %%xmm1, %0\n\t" 
				:"=m" (c)
				:"r" (a), "r" (b)	
					
			);
			
        
    		}
		
	
		//stop the timer
		gettimeofday(&_tstop, &tz);

		//get seconds from the structures
		t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
		t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
		//compute the elapsed time
		ElapsedTime = (float)(t2 - t1);

		//display the result
		printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
		printf("%s%g%s","Dot Product = ",c[0],"\n\n");
	}
	
	else{
	
		printf("SSE instructions not available on this CPU\n\n");
		
	}
	
	
}//end fnC()

//entry function
int main()
{
	//local declarations
	//integers
	int bTERM;		//menu flag

	//character strings
	char inchar;		//input character
	

	//initialize bTERM to NOT TERMINATE
	bTERM = 0;
	
	//clear terminal
	system("clear");
	
	do {
		
		//present menu to user
		printf("\nEnter the Number of the routine to run, and press <Enter> \n");
		printf("1\t4-D Dot Product using x87 instructions\n");
		printf("2\t4-D Dot Product using 3dNow! SIMD Instructions\n");
		printf("3\t4-D Dot Product using SSE SIMD Instructions\n");
		printf("q\tEnd this program\n");
	
		//get user input
		scanf("%s", &inchar);

		//call the user selected function
		switch	(inchar){

			case '1': fnA();
				break;
	
			case '2': fnB();
				break;
				
			case '3': fnC();
				break;
		
			//select the terminate function	
			case 'q': printf("Ending This Program\n");
				
				//set flag to terminate
				bTERM = 1;
				break;

		}
		
	} while(!bTERM);

	return 0;

} // End Main

