// Par03.c : implementation file
//

// This program compares the numerical integration using straight x87 floating
// point code to use of 3dNow! and SSE SIMD code.  If the run-time processor
// does not support SIMD, the SIMD routine simply exits with a 
// message (CPUid MUST be supported, however...I did not test for
// that support before making the calls to CPUid).
//
// The x87 routine is in fnA().
// The 3dNow! routine is in fnB().
// The SSE routine is in fnC()


#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

//defines
#define NumIts 50000		//number of iterations

//functions
void fnA();			
void fnB();
void fnC();			


// perform Simpson's Rule Integration using straight x87 code.
void fnA()
{

	//general local declarations
	//integers
	int counter;		// iteration counter
	int SimpCounter;	//simpson's rule loop counter
	int ItCounter;		//iteration counter
	
	//floats
	float h_3;			//simpson's rule parameters
	float S0;
	float S1;
	float S2;
	float Integral;			//result
	float f[10000];			//function f(x)
	
	//doubles
	double argX;			//x for x^5 assignment
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//assign the function f(x) = x^5
	for(counter=0; counter < 10000; ++counter){

		argX = (double)counter * 0.0001;
		f[counter] = (float)pow(argX, 5);

	}


	//get start time
	gettimeofday(&_tstart, &tz);

	//h/3, where h=(b-1)/n
	//moved out of loop to be consistent with SIMD routine
	h_3 = (float)0.999 /30000;

	for(ItCounter=1; ItCounter <= NumIts; ++ItCounter){

		//sum the first and last terms
		S0 = f[0] + f[9999];

		//zero the values for the running summations
		S1 = 0;
		S2 = 0;

		//do the unrolled summation
		for(SimpCounter = 1; SimpCounter <= 9969; SimpCounter = SimpCounter + 16){

			S1 += f[SimpCounter] + f[SimpCounter + 2] + f[SimpCounter + 4] + f[SimpCounter + 6] + f[SimpCounter + 8]
					+ f[SimpCounter + 10] + f[SimpCounter + 12] + f[SimpCounter + 14];
			S2 += f[SimpCounter + 1] + f[SimpCounter + 3] + f[SimpCounter + 5] + f[SimpCounter + 7] + f[SimpCounter + 9]
					+ f[SimpCounter + 11] + f[SimpCounter + 13] + f[SimpCounter + 15];

		}

		//clean up extra terms
		S1 += f[9985] + f[9987] + f[9989] + f[9991] + f[9993] + f[9995] + f[9997];
		S2 += f[9986] + f[9988] + f[9990] + f[9992] + f[9994] + f[9996] + f[9998];

		// 'odd; terms * 2
		S1 *= 2;

		// 'even' terms * 4
		S2 *= 4;

		// total summation
		Integral = S0 + S1 + S2;
		
		//multiply by h/3
		Integral *= h_3;

	}
	
	//stop the timer
	gettimeofday(&_tstop, &tz);

	//get seconds from the structures
	t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
	t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
	//compute the elapsed time
	ElapsedTime = (float)(t2 - t1);

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	printf("%s%g%s", "Integral = ",Integral,"\n");
	
}//end fnA()

// perform Simpson's Rule integration in ASSEMBLY using 3dNow! SIMD instructions
void fnB()
{

	//general local declarations
	//integers
	int counter;		//iteration counter
	int bAMDOK;		//cpu supports 3DNow! instructions
	int SimpCounter;	//simpson's rule loop counter
	int ItCounter;		//iteration counter
	
	//floats
	float h_3;		//simpson's rule parameters
	float Integral;		//result
	float f[10000];		//function f(x)
	float factvec[2];	//vector for Simpson's Rule factors
	float tempval;		//holder for reordering f[x] values
	
	//doubles
	double argX;		//x for x^5 assignment
	
	//variables for time execution time
	//floats
	float ElapsedTime;	//ticks during loop
	
	//doubles
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//initialize AMD flag to zero
	bAMDOK = 0;
	
	//initialize factor vector
	factvec[0] = 4;
	factvec[1] = 2;

	//test for 3DNow! capability in cpu
	//this is adapted from code in AMD 3DNow! Tech. Manual
	asm volatile(
		"pushl	%%eax\n\t"
		"pushl	%%edx\n\t"
		"xor	%%eax, %%eax\n\t"
		"CPUid\n\t"
		"test	%%eax, %%eax\n\t"
		"jz	No3dNow\n\t"	
		"movl	$0x80000000, %%eax\n\t"
		"CPUid\n\t"
		"cmp	$0x80000001, %%eax\n\t"
		"jl	No3dNow\n\t"
		"movl	$0x80000001, %%eax\n\t"
		"CPUid\n\t"
		"test	$0x80000000, %%edx\n\t"
		"jnz	Yes3dNow\n\t"
		"jmp	No3dNow\n"
"Yes3dNow:	\n\t"
		"movl	$1, %0\n\t"
		"jmp	EndCpuTest\n"
"No3dNow:	\n\t" 
		"movl	$0,%0\n"
"EndCpuTest:	\n\t"	 
		"popl	%%edx\n\t"
		"popl	%%eax"
		:"=g" (bAMDOK)
		:
		:"%eax","%ebx","ecx","%edx"
		
	);

	if(bAMDOK){

		//display entry message
		printf("\nNumerical Integration using 3dNow! SIMD Code...\n");
	
		//assign the function f(x) = x^5
		for(counter=0; counter < 10000; ++counter){

			argX = (double)counter * 0.0001;
			f[counter] = (float)pow(argX, 5);

		}
		
		//get start time
		gettimeofday(&_tstart, &tz);
		
		//switch f[1] and f[n]
		//this is so f[0] and f[n] are in sequential bytes
		//to be accessed as a vector
		tempval = f[1];
		f[1] = f[9999];
		f[9999] = tempval;
		
		//h/3, where h=(b-a)/n
		//moved out of loop to avoid mmx vs fp register collision
		h_3 = (float)0.999 /30000;
		
		//switch fp registers to SIMD mode
		asm ("femms\n\t");
		
	
		//begin iterative integrations 
		for(ItCounter=1; ItCounter <= NumIts; ++ItCounter){ 
		
			
			
			asm volatile(
		
				//initialize mm0 with f[2]|f[3]
				"movq	8(%0), %%mm0\n\t"
				
				//mm3 = h_3|h_3 for use after the summation
				"movd	  %1, %%mm3\n\t"
				"punpckldq %%mm3, %%mm3\n\t"
				
				//zero the accumulation registers
				"pxor   %%mm4, %%mm4\n\t"
				"pxor   %%mm5, %%mm5\n\t"
				"pxor   %%mm6, %%mm6\n\t"
				"pxor   %%mm7, %%mm7\n\t"
				:
				:"r" (f), "r" (h_3)
			
			);
		
			//do the partially unrolled summation		
			for(SimpCounter = 4; SimpCounter <= 9972; SimpCounter = SimpCounter + 16){

				asm volatile(

					//prefetch data for this iteration
					//of loop
					"prefetch 512(%1,%0,4)\n\t"
					
					//accumulate into available registers
					//use them all to avoid register
					//dependency delay
					"pfadd	  (%1,%0,4), %%mm0\n\t"
					"pfadd	 8(%1,%0,4), %%mm4\n\t"
					"pfadd	16(%1,%0,4), %%mm0\n\t"
					"pfadd	24(%1,%0,4), %%mm5\n\t"
					"pfadd	32(%1,%0,4), %%mm0\n\t"
					"pfadd	40(%1,%0,4), %%mm6\n\t"
					"pfadd	48(%1,%0,4), %%mm0\n\t"
					"pfadd	56(%1,%0,4), %%mm7\n\t"
					:
					:"r" (SimpCounter), "r" (f)
					
				
				);

			}
		
			asm volatile(
			
				//begin prefetching data from mem into cache
				"prefetchnta 384(%1,%2,4)\n\t"
				
				//clean up terms not in sum
				//again using available registers to avoid
				//register dependency delay
				"pfadd	  (%1,%2,4), %%mm0\n\t"
				"pfadd	 8(%1,%2,4), %%mm4\n\t"
				"pfadd	16(%1,%2,4), %%mm5\n\t"
				"pfadd	24(%1,%2,4), %%mm0\n\t"
				"pfadd	32(%1,%2,4), %%mm6\n\t"
				"pfadd	40(%1,%2,4), %%mm7\n\t"
				
				//accumulate into mm0
				"pfadd  %%mm7, %%mm5\n\t"
				"pfadd  %%mm4, %%mm0\n\t"
				"pfadd  %%mm5, %%mm6\n\t"
				"pfadd  %%mm6, %%mm0\n\t"
		
			//compute products and complete sums of S0, S1, S2
		
				"movq	(%3), %%mm1\n\t"
				"movq	(%1), %%mm2\n\t"
				"pfmul	%%mm1, %%mm0\n\t"
				"pfadd	%%mm2, %%mm0\n\t"
				"pfmul	%%mm3, %%mm0\n\t"
				"pfacc	%%mm0, %%mm0\n\t"
				"movd	%%mm0, %0\n\t"
				:"=m" (Integral)
				:"r" (f), "r" (0x2704), "r" (factvec)
		
			);

		}

		//clear fp registers from SIMD mode
		asm ("femms\n\t");
	
		//stop the timer
		gettimeofday(&_tstop, &tz);

		//get seconds from the structures
		t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
		t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
		//compute the elapsed time
		ElapsedTime = (float)(t2 - t1);

		//display the result
		printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
		printf("%s%g%s", "Integral = ",Integral,"\n");
	
	}
	
	else{
	
		printf("3dNow instructions not available on this CPU\n\n");
		
	}


}//end fnB()

// perform Simpson's Rule integration in ASSEMBLY using SSE SIMD instructions
void fnC()
{

	//general local declarations
	//integers
	int counter;		//iteration counter
	int bSSEOK;		//cpu supports SSE instructions
	int SimpCounter;	//simpson's rule loop counter
	int ItCounter;		//iteration counter
	
	//floats	
	float h_3;		//simpson's rule parameters
	float Integral;		//result
	float partialvec[4];	//1/2qword=4,2, 1/2qword=1
	float f[10000];		//function f(x)
	float factvec[4];	//vector for Simpson's Rule factors
	float zerobytes[4];	//some zero bytes to clear registers
	float tempval;		//holder for reordering f[x] values	
	
	//doubles	
	double argX;		//x for x^5 assignment		
		
	//variables for time execution time
	//floats	
	float ElapsedTime;	//ticks during loop	
	
	//doubles	
	double t1;		//ticks at start of loop
	double t2;		//ticks at loop end		
	
	//structures needed for basic timing functions
	static struct timeval _tstart, _tstop;
	static struct timezone tz;

	//initialize SSE flag to zero
	bSSEOK = 0;
	
	//initialize factor vector
	factvec[0] = 4;
	factvec[1] = 2;
	factvec[2] = 4;
	factvec[3] = 2;
	
	//initialize partialh_3
	partialvec[0] = 1;
	partialvec[1] = 1;
	partialvec[2] = 4;
	partialvec[3] = 2;
	
	//initialize the zero bytes
	zerobytes[0] = 0;
	zerobytes[1] = 0;
	zerobytes[2] = 0;
	zerobytes[3] = 0;
	
	//test for SSE capability in cpu
	asm volatile(
	
		"pushl	%%eax\n\t"
		"pushl	%%edx\n\t"
		"xor	%%eax, %%eax\n\t"
		"CPUid\n\t"
		"test	%%eax, %%eax\n\t"
		"jz	NoSSE\n\t"	
		"movl	$0x00000001, %%eax\n\t"
		"CPUid\n\t"
		"test	$0x02000000, %%edx\n\t"
		"jnz	YesSSE\n\t"
		"jmp	NoSSE\n"
"YesSSE:	\n\t"
		"movl	$1, %0\n\t"
		"jmp	EndP3Test\n"
"NoSSE	:	\n\t" 
		"movl	$0,%0\n"
"EndP3Test:	\n\t"	 
		"popl	%%edx\n\t"
		"popl	%%eax"
		:"=g" (bSSEOK)
		:
		:"%eax","%ebx","%ecx","%edx"
		
	);

	if(bSSEOK){

		//display entry message
		printf("\nNumerical Integration using SSE SIMD Code...\n");
		
		//assign the function f(x) = x^5
		for(counter=0; counter < 10000; ++counter){

			argX = (double)counter * 0.0001;
			f[counter] = (float)pow(argX, 5);

		}
		
		//initialize result to zero
		Integral = 0;
		
		//get start time
		gettimeofday(&_tstart, &tz);
	
		//switch f[1] and f[n]
		//this is so f[0] and f[n] are in sequential bytes
		//to be accessed as a vector
		tempval = f[1];
		f[1] = f[9999];
		f[9999] = tempval;
		
		//h/3, where h=(b-a)/n
		//moved out of loop to avoid mmx vs fp register collision
		h_3 = (float)0.999 /30000;

		//begin iterative integrations 
		for(ItCounter=1; ItCounter <= NumIts; ++ItCounter){ 
			
			
			asm volatile(
		
				//preload load f[4]..f[7] into xmm0 
				"movaps   16(%0), %%xmm0\n\t"
				
				//broadcast h_x into xmm3
				"movss    %1, %%xmm3\n\t"
				"shufps $0x00, %%xmm3, %%xmm3\n\t"
				
				//load h_3 into high dword of xmm7
				//1 into low dword of xmm7
				"movaps    (%2), %%xmm7\n\t"
				
				//zero accumulation registers
				//pxor failed on AthlonXP
				"movaps    (%3), %%xmm4\n\t"
				"movaps    (%3), %%xmm5\n\t"
				"movaps    (%3), %%xmm6\n\t"
				: //no outputs//
				:"r" (f), "m" (h_3), "r" (partialvec), "r"
				(zerobytes)
			
			);
		
			//do the partially unrolled summation		
			for(SimpCounter = 8; SimpCounter <= 9976; SimpCounter = SimpCounter + 16){

				asm volatile(

					"addps   (%1,%0,4), %%xmm0\n\t"
					"addps 16(%1,%0,4), %%xmm4\n\t"
					"addps 32(%1,%0,4), %%xmm5\n\t"
					"addps 48(%1,%0,4), %%xmm6\n\t"
					://no output registers//
					:"r" (SimpCounter), "r" (f)
				);

			}
		
			asm volatile(
				
				//clean up terms not in sum
				"addps	  (%1,%2,4), %%xmm0\n\t"
				"addps	16(%1,%2,4), %%xmm5\n\t"
								
				//accumulate into mm0
				"addps  %%xmm4, %%xmm0\n\t"
				"addps  %%xmm5, %%xmm6\n\t"
				"addps  %%xmm6, %%xmm0\n\t"
		
			//compute products and complete sums of S0, S1, S2
		
				"movaps   (%1), %%xmm2\n\t"
				"movaps	  (%3), %%xmm4\n\t"
				"mulps  %%xmm4, %%xmm0\n\t"
				"mulps  %%xmm7, %%xmm2\n\t" 
				"addps  %%xmm2, %%xmm0\n\t"
				"mulps  %%xmm3, %%xmm0\n\t"
				
				//accumulate xmm0 to a scalar result
				"movaps %%xmm0, %%xmm1\n\t"
				"shufps $0xee, %%xmm0, %%xmm1\n\t"
				"addps  %%xmm0, %%xmm1\n\t"
				"movaps %%xmm1, %%xmm0\n\t"
				"shufps $0xe5, %%xmm1, %%xmm1\n\t"
				"addps  %%xmm0, %%xmm1\n\t"
				"movss  %%xmm1, %0\n\t"
				:"=m" (Integral)
				:"r" (f), "r" (0x2708), "r" (factvec)
		
			);

		}
		

		//stop the timer
		gettimeofday(&_tstop, &tz);

		//get seconds from the structures
		t1 = (double)_tstart.tv_sec + (double)_tstart.tv_usec * 0.000001;
		t2 = (double)_tstop.tv_sec + (double)_tstop.tv_usec * 0.000001;
	
		//compute the elapsed time
		ElapsedTime = (float)(t2 - t1);

		//display the result
		printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
		printf("%s%g%s","Integral = ",Integral,"\n\n");
	}
	
	else{
	
		printf("SSE instructions not available on this CPU\n\n");
		
	}
	
	
}//end fnC()

//entry function
int main()
{
	//local declarations
	//integers
	int bTERM;		//menu flag

	//character strings
	char inchar;		//input character
	

	//initialize bTERM to NOT TERMINATE
	bTERM = 0;
	
	//clear terminal
	system("clear");
	
	do {
		
		//present menu to user
		printf("\nEnter the Number of the routine to run, and press <Enter> \n");
		printf("1\tSimpson's Rule Integration using x87 instructions\n");
		printf("2\tSimpson's Rule Integration using 3dNow! SIMD Instructions\n");
		printf("3\tSimpson's Rule Integration using SSE SIMD Instructions\n");
		printf("q\tEnd this program\n");
	
		//get user input
		scanf("%s", &inchar);

		//call the user selected function
		switch	(inchar){

			case '1': fnA();
				break;
	
			case '2': fnB();
				break;
				
			case '3': fnC();
				break;
		
			//select the terminate function	
			case 'q': printf("Ending This Program\n");
				
				//set flag to terminate
				bTERM = 1;
				break;

		}
		
	} while(!bTERM);

	return 0;

} // End Main

