// Par01.c : implementation file
//
// This program compares the dot product using straight x87 floating
// point code to use of SIMD code.  If the run-time processor
// does not support SIMD, the SIMD routines simply exit with a 
// message (CPUid MUST be supported, however...I did not test for
// that support before making the calls to CPUid).
//
// The x87 routine is in fnA().
// The 3dNow! routine is in fnB().
// The SSE routine is in fnC().

#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <stdlib.h>

//defines
#define NumIts 100000000		//number of iterations

//functions
void fnA();			
void fnB();			

// perform dot product using x87 code as compiled
// from C unrolled loop as demo Loop02.
void fnA()
{

	//general local declarations
	//integers
	int counter;			  // iteration counter

	//floats
	float a[3];		//vector a
	float b[3];		//vector b
	float c;		//dot product result
	
	//variables for time execution time
	//doubles
	double starttime;		//system ticks at start of loop
	double stoptime;		//system ticks at loop exit
	double ElapsedTime;		//system ticks during loop

	//initialize the two vectors with some made up values
	a[0] = (float)23.45;
	a[1] = (float)14.65;
	a[2] = (float)33.87;

	b[0] = (float)36.99;
	b[1] = (float)21.32;
	b[2] = (float)17.38;
		
	//display entry message
	printf("\nComputing Unrolled Dot Product using only x87 ...\n");
	
	//get start time
	starttime = GetTickCount();
	
	for(counter=1; counter <= NumIts; ++counter){
    
		c = a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
		
   	}
	
	//stop the timer
	stoptime = GetTickCount();
	
	//compute the elapsed time
	ElapsedTime = (stoptime - starttime) * 0.001;

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	printf("%s%g%s", "Dot product = ",c,"\n");
	
}//end fnA()

// perform dot product in ASSEMBLY using 3dNow! SIMD instructions
void fnB()
{

	//general local declarations
	//integers
	int counter;			//iteration counter
	int bAMDOK;			//cpu supports 3DNow! instructions

	//floats
	float a[4];			//vector a, 4-d
	float b[4];			//vector b, 4-d
	float c;			//dot product result
	
	//variables for time execution time
	//doubles
	double starttime;		//system ticks at start of loop
	double stoptime;		//system ticks at loop exit
	double ElapsedTime;		//system ticks during loop

	//initialize AMD flag to zero
	bAMDOK = 0;

	//test for 3DNow! capability in cpu
	//this is adapted from code in AMD 3DNow! Tech. Manual
	_asm {

			push	eax
			push	edx
			xor		eax, eax		; CPUid function Vendor ID
			CPUid
			test	eax, eax
			jz		$No3dNow		
			mov		eax, 80000000h	; CPUid function Largest Ext. Val
			CPUid
			cmp		eax, 80000001h
			jl		$No3dNow
			mov		eax, 80000001h	; CPUid function Sig. + features
			CPUid
			test	edx, 80000000h
			jnz		$Yes3dNow
			jmp		$No3dNow
$Yes3dNow:	mov		bAMDOK, 1		; bAMDOK = true
			jmp		$EndCpuTest
$No3dNow:	mov		bAMDOK, 0		; bAMDOK = false
$EndCpuTest:		pop		edx
			pop		eax

	}

	if(bAMDOK){

		//initialize the two vectors with some made up values
		a[0] = (float)23.45;
		a[1] = (float)14.65;
		a[2] = (float)33.87;
		a[3] = (float)0;

		b[0] = (float)36.99;
		b[1] = (float)21.32;
		b[2] = (float)17.38;
		b[3] = (float)0;

		//start the timer
		starttime = GetTickCount(); 	

		//switch registers to mmx mode
		_asm	femms

		for(counter=1; counter <= NumIts; ++counter){

			
			//3DNow! dot product, have to use assembly
			//technically, we are doing 4-dimension vectors
			//here to avoid the movd instruction
			//also recall that [index] in ASSEMBLY refers to
			//actual byte offsets.
			__asm {

				movq		mm0, a[0]		; mm(0) = a(0) | a(1)
				movq		mm3, b[0]		; mm(1) = b(0) | b(1)
				pfmul		mm0, mm3		; mm(0) = a(0)*b(0) | a(1) * b(1)
				movq		mm2, a[8]		; don't use movd!!	SLOW!
				movq		mm1, b[8]
				pfacc		mm0, mm0		
				pfmul		mm1, mm2
				pfadd		mm0, mm1		
				movd		c, mm0			; c = mm(0)lo
			
			
			}

		}

		//switch registers back to x87 mode
		_asm	femms
	
		//stop the timer
		stoptime = GetTickCount();		
	
		//compute the elapsed time, in seconds
		ElapsedTime = (stoptime - starttime) * .001;

		//display the result in a message box
		printf("Using 3DNow!, performed dot product 100,000,000 times in %g", ElapsedTime);
		printf("%s", " seconds.\n");
		printf("%s%g", "Result = ",c);

		
	}
		
	else {

		printf("This processor does not support 3dNow! Instructions\n");
			
	}
	
	
}//end fnB()

// perform dot product in ASSEMBLY using SSE SIMD instructions
void fnC()
{

	//general local declarations
	//integers
	int counter;		//iteration counter
	int bSSEOK;			//cpu supports SSE instructions

	//floats
	float a[4];			//vector a, 4-d
	float b[4];			//vector b, 4-d
	float c;			//dot product result
	
	//variables for time execution time
	//doubles
	double starttime;		//system ticks at start of loop
	double stoptime;		//system ticks at loop exit
	double ElapsedTime;		//system ticks during loop

	//initialize SSE flag to zero
	bSSEOK = 0;

	//test for SSE capability in cpu
	_asm {

			push	eax
			push	edx
			xor		eax, eax		; CPUid function Vendor ID
			CPUid
			test	eax, eax
			jz		$NoSSE		
			mov		eax, 00000001h	; CPUid function Largest Ext. Val
			CPUid
			test	eax, 02000000h
			jl		$YesSSE
			jmp		$NoSSE
$YesSSE:	mov		bSSEOK, 1		; bSSEOK = true
			jmp		$EndCpuTest
$NoSSE:		mov		bSSEOK, 0		; bSSEOK = false
$EndCpuTest:		pop		edx
			pop		eax

	}

	if(bSSEOK){

		//initialize the two vectors with some made up values
		a[0] = (float)23.45;
		a[1] = (float)14.65;
		a[2] = (float)33.87;
		a[3] = (float)0;

		b[0] = (float)36.99;
		b[1] = (float)21.32;
		b[2] = (float)17.38;
		b[3] = (float)0;

		//start the timer
		starttime = GetTickCount(); 	

		//switch registers to mmx mode
		_asm	femms

		for(counter=1; counter <= NumIts; ++counter){

			
			//SSE dot product, have to use assembly
			//technically, we are doing 4-dimension vectors
			//to make full use of 128 bit SSE registers
			//in essence, we get the 4th d for free
			__asm {

				movaps		xmm0, a[0]
				movaps		xmm1, b[0]
				mulps		xmm0, xmm1
				movaps		xmm1, xmm0
				shufps		xmm1, xmm0, 0xee
				addps		xmm1, xmm0
				movaps		xmm0, xmm1
				shufps		xmm1, xmm1, 0xe5
				addps		xmm1, xmm0
				movss		xmm0, c
			
			
			}

		}

		//switch registers back to x87 mode
		_asm	femms
	
		//stop the timer
		stoptime = GetTickCount();		
	
		//compute the elapsed time, in seconds
		ElapsedTime = (stoptime - starttime) * .001;

		//display the result in a message box
		printf("Using SSE, performed dot product 100,000,000 times in %g", ElapsedTime);
		printf("%s", " seconds.\n");
		printf("%s%g", "Result = ",c);

		
	}
		
	else {

		printf("This processor does not support SSE Instructions\n");
			
	}
	
	
}//end fnC()

//entry function
int main()
{
	//local declarations
	//integers
	int bTERM;		//menu flag

	//character strings
	char inchar;		//input character
	

	//initialize bTERM to NOT TERMINATE
	bTERM = 0;
	
	//clear terminal
	system("cls");
	
	do {
		
		//present menu to user
		printf("\nEnter the Number of the routine to run, and press <Enter> \n");
		printf("1\t4-D Dot Product using x87 instructions\n");
		printf("2\t4-D Dot Product using 3dNow! SIMD Instructions\n");
		printf("3\t4-D Dot Product using SSE SIMD Instructions\n");
		printf("q\tEnd this program\n");
	
		//get user input
		scanf("%s", &inchar);

		//call the user selected function
		switch	(inchar){

			case '1': fnA();
				break;
	
			case '2': fnB();
				break;

			case '3': fnC();
				break;
		
			//select the terminate function	
			case 'q': printf("Ending This Program\n");
				
				//set flag to terminate
				bTERM = 1;
				break;

		}
		
	} while(!bTERM);

	return 0;

} // End Main

