// Par03.c : implementation file
//

// This program compares the numerical integration using straight x87 floating
// point code to use of 3dNow! and SSE SIMD code.  If the run-time processor
// does not support SIMD, the SIMD routine simply exits with a 
// message (CPUid MUST be supported, however...I did not test for
// that support before making the calls to CPUid).
//
// The x87 routine is in fnA().
// The 3dNow! routine is in fnB().
// The SSE routine is in fnC()

#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

//defines
#define cNumIts 50000		//number of iterations

//functions
void fnA();			
void fnB();
void fnC();			

// perform Simpson's Rule Integration using straight x87 code.
void fnA()
{

	//local declarations
	//integers
	int SimpCounter;		//simpson's rule loop counter
	int ItCounter;			//iteration counter
	int Counter;			//counter to assign function

	//floats
	float h_3;				//simpson's rule parameters
	float S0;
	float S1;
	float S2;
	float Integral;			//result
	float f[10000];			//function f(x)
	
	//doubles
	double argX;			//x for x^5 assignment
	double starttime;		//ticks at start of main loop
	double stoptime;		//ticks at main loop exit
	double ElapsedTime;		//ticks during loop

	//assign the function f(x) = x^5
	for(Counter=0; Counter < 10000; ++Counter){

		argX = (double)Counter * 0.0001;
		f[Counter] = (float)pow(argX, 5);

	}

	printf("\nStarting Simpson's Rule Integration using x87 code ...\n");
	
	//start the timer
	starttime = GetTickCount();

	//h/3, where h=(b-1)/n
	//moved out of loop to be consistent with SIMD routine
	h_3 = (float)0.999 /30000;

	for(ItCounter=1; ItCounter <= cNumIts; ++ItCounter){

		//sum the first and last terms
		S0 = f[0] + f[9999];

		//zero the values for the running summations
		S1 = 0;
		S2 = 0;

		//do the unrolled summation
		for(SimpCounter = 1; SimpCounter <= 9969; SimpCounter = SimpCounter + 16){

			S1 += f[SimpCounter] + f[SimpCounter + 2] + f[SimpCounter + 4] + f[SimpCounter + 6] + f[SimpCounter + 8]
					+ f[SimpCounter + 10] + f[SimpCounter + 12] + f[SimpCounter + 14];
			S2 += f[SimpCounter + 1] + f[SimpCounter + 3] + f[SimpCounter + 5] + f[SimpCounter + 7] + f[SimpCounter + 9]
					+ f[SimpCounter + 11] + f[SimpCounter + 13] + f[SimpCounter + 15];

		}

		//clean up extra terms
		S1 += f[9985] + f[9987] + f[9989] + f[9991] + f[9993] + f[9995] + f[9997];
		S2 += f[9986] + f[9988] + f[9990] + f[9992] + f[9994] + f[9996] + f[9998];

		// 'odd; terms * 2
		S1 *= 2;

		// 'even' terms * 4
		S2 *= 4;

		// total summation
		Integral = S0 + S1 + S2;
		
		//multiply by h/3
		Integral *= h_3;

	}

	//stop the timer
	stoptime = GetTickCount();

	//compute the elapsed time
	ElapsedTime = (stoptime - starttime) * 0.001;

	//display the result
	printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
	printf("%s%g%s", "Integral = ", Integral, "\n");
	
}//end fnA()

// perform Simpson's Rule integration in ASSEMBLY using 3dNow! SIMD instructions
void fnB()
{

	//integers
	int SimpCounter;		//simpson's rule loop counter
	int ItCounter;			//iteration counter
	int Counter;			//counter to assign function
	int bAMDOK;			//cpu supports 3DNow! instructions
	
	//floats
	float h_3;				//simpson's rule parameters
	float Integral;			//result
	float f[10000];			//function f(x)
	float tempval;			//temp for switching terms
	float factvec[2];		//factors 4 and 2
	
	//doubles
	double argX;			//x for x^5 assignment
	double starttime;		//ticks at start of main loop
	double stoptime;		//ticks at main loop exit
	double ElapsedTime;		//ticks during loop

	printf("\nStarting Simpson's Rule Integration using 3dNow! code ...\n");

	//test for 3DNow! capability in cpu
	//this is adapted from code in AMD 3DNow! Tech. Manual
	_asm {

				push	eax
				push	edx
				xor		eax, eax		; CPUid function Vendor ID
				CPUid
				test	eax, eax
				jz		$No3dNow		
				mov		eax, 80000000h	; CPUid function Largest Ext. Val
				CPUid
				cmp		eax, 80000001h
				jl		$No3dNow
				mov		eax, 80000001h	; CPUid function Sig. + features
				CPUid
				test	edx, 80000000h
				jnz		$Yes3dNow
				jmp		$No3dNow
$Yes3dNow:		mov		bAMDOK, 1		; bAMDOK = true
				jmp		$EndCpuTest
$No3dNow:		mov		bAMDOK, 0		; bAMDOK = false
$EndCpuTest:	pop		edx
				pop		eax

	}

	if(bAMDOK){

	//assign the function f(x) = x^5
	for(Counter=0; Counter < 10000; ++Counter){

		argX = (double)Counter * 0.0001;
		f[Counter] = (float)pow(argX, 5);

	}

	//initialize the factors
	factvec[0] = 4;
	factvec[1] = 2;

	//start the timer
	starttime = GetTickCount();

	//switch f[1] and f[n]
	//this is needed to access first and last as a single
	//eight byte 'vector'
	tempval = f[1];
	f[1] = f[9999];
	f[9999] = tempval;

	//h/3, where h=(b-1)/n
	//moved out of loop to avoid mmx vs fp register collision
	h_3 = (float)0.999 /30000;

	_asm femms

	for(ItCounter=1; ItCounter <= cNumIts; ++ItCounter){

		__asm {	
		
				;preload mmo with f[3]|f[2]
				movq	mm0, f[8]

				;preload and broadcast h_3 into mm3
				movd	mm3, h_3
				punpckldq	mm3, mm3

				;save eax
				push	eax

				;zero accumulation registers
				pxor	mm4, mm4
				pxor	mm5, mm5
				pxor	mm6, mm6
				pxor	mm7, mm7
		
		}



		//do the unrolled summation
		for(SimpCounter = 4; SimpCounter <= 9972; SimpCounter = SimpCounter + 16){

			__asm{

				;get current value of SimpCounter for
				;indexing
				mov		eax, SimpCounter

				;begin prefetching data for this iteration
				prefetch		f[512 + eax * 4]

				;accumulate into available registers
				;use them all to avoid register dependency delay
				pfadd	mm0, f[eax * 4]
				pfadd	mm4, f[8 + eax * 4] 
				pfadd	mm0, f[16 + eax * 4] 
				pfadd	mm5, f[24 + eax * 4] 
				pfadd	mm0, f[32 + eax * 4] 
				pfadd	mm6, f[40 + eax * 4]
				pfadd	mm0, f[48 + eax * 4]
				pfadd	mm7, f[56 + eax * 4]
				
			}

		}

		
		_asm{

				;begin prefetching data from mem
				prefetchnta		f[384 + 0x2704 * 4]

				;clean up terms not in sum
				;using available registers to avoid
				;dependencies
				pfadd   mm0, f[0 + 0x2704 * 4]
				pop		eax
				pfadd   mm4, f[8 + 0x2704 * 4]
				pfadd   mm5, f[16 + 0x2704 * 4]
				pfadd	mm0, f[24 + 0x2704 * 4]
				pfadd	mm6, f[32 + 0x2704 * 4]
				pfadd	mm7, f[40 + 0x2704 * 4]

				;accumulate into mm0
				pfadd	mm5, mm7
				pfadd	mm0, mm4
				pfadd	mm6, mm5
				pfadd	mm0, mm6

				;compute products and complete sums
				;of S0, S1, S2
				movq	mm1, factvec[0]
				movq	mm2, f[0]
				pfmul	mm0, mm1
				pfadd	mm0, mm2
				pfmul	mm0, mm3
				pfacc	mm0, mm0
				movd	Integral, mm0
		
		}
		

	}

		_asm femms

		//stop the timer
		stoptime = GetTickCount();		
	
		//compute the elapsed time, in seconds
		ElapsedTime = (stoptime - starttime) * .001;

		//display the result in a message box
		printf("Using 3DNow!, Integrated x^5 50,000 times in %g", ElapsedTime);
		printf("%s", " seconds.\n");
		printf("%s%g%s", "Integral = ", Integral, "\n");
		
	}
		
	else {

		printf("This processor does not support 3dNow! Instructions\n");
			
	}
		
}//end fnB()

// perform Simpson's Rule integration in ASSEMBLY using SSE SIMD instructions
void fnC()
{

	//general local declarations
	//integers
	int counter;		//iteration counter
	int bSSEOK;		//cpu supports SSE instructions
	int SimpCounter;	//simpson's rule loop counter
	int ItCounter;		//iteration counter
	
	//floats	
	float h_3;		//simpson's rule parameters
	float Integral;		//result
	float partialvec[4];	//1/2qword=4,2, 1/2qword=1
	float f[10000];		//function f(x)
	float factvec[4];	//vector for Simpson's Rule factors
	float zerobytes[4];	//some zero bytes to clear registers
	float tempval;		//holder for reordering f[x] values	
	
	//doubles	
	double argX;		//x for x^5 assignment		
	double starttime;		//ticks at start of main loop
	double stoptime;		//ticks at main loop exit
	double ElapsedTime;		//ticks during loop
		
	//initialize SSE flag to zero
	bSSEOK = 0;
	
	//initialize factor vector
	factvec[0] = 4;
	factvec[1] = 2;
	factvec[2] = 4;
	factvec[3] = 2;
	
	//initialize partialh_3
	partialvec[0] = 1;
	partialvec[1] = 1;
	partialvec[2] = 4;
	partialvec[3] = 2;
	
	//initialize the zero bytes
	zerobytes[0] = 0;
	zerobytes[1] = 0;
	zerobytes[2] = 0;
	zerobytes[3] = 0;
	
	//test for SSE capability in cpu
	_asm {

			push	eax
			push	edx
			xor		eax, eax		; CPUid function Vendor ID
			CPUid
			test	eax, eax
			jz		$NoSSE		
			mov		eax, 00000001h	; CPUid function Largest Ext. Val
			CPUid
			test	eax, 02000000h
			jl		$YesSSE
			jmp		$NoSSE
$YesSSE:	mov		bSSEOK, 1		; bSSEOK = true
			jmp		$EndCpuTest
$NoSSE:		mov		bSSEOK, 0		; bSSEOK = false
$EndCpuTest:		pop		edx
			pop		eax

	}

	if(bSSEOK){

		//display entry message
		printf("\nNumerical Integration using SSE SIMD Code...\n");
		
		//assign the function f(x) = x^5
		for(counter=0; counter < 10000; ++counter){

			argX = (double)counter * 0.0001;
			f[counter] = (float)pow(argX, 5);

		}
		
		//initialize result to zero
		Integral = 0;
		
		//start the timer
		starttime = GetTickCount();
	
		//switch f[1] and f[n]
		//this is so f[0] and f[n] are in sequential bytes
		//to be accessed as a vector
		tempval = f[1];
		f[1] = f[9999];
		f[9999] = tempval;
		
		//h/3, where h=(b-a)/n
		//moved out of loop to avoid mmx vs fp register collision
		h_3 = (float)0.999 /30000;

		//begin iterative integrations 
		for(ItCounter=1; ItCounter <= cNumIts; ++ItCounter){ 
			
			
			_asm {
		
				;preload load f[4]..f[7] into xmm0 
				movaps   xmm0, f[16]
				
				;broadcast h_x into xmm3
				movss    xmm3, h_3
				shufps   xmm3, xmm3, 0x00
				
				;load h_3 into high dword of xmm7
				;1 into low dword of xmm7
				movaps    xmm7, partialvec
				
				;save eax
				push eax

				;zero accumulation registers
				movaps    xmm4, zerobytes[0]
				movaps    xmm5, zerobytes[0]
				movaps    xmm6, zerobytes[0]
			
			}
		
			//do the partially unrolled summation		
			for(SimpCounter = 8; SimpCounter <= 9976; SimpCounter = SimpCounter + 16){

				_asm {


					;accumulate sum for this iteration
					addps   xmm0, f[eax * 4]
					addps   xmm4, f[16 + eax * 4]
					addps   xmm5, f[32 + eax * 4]
					addps   xmm6, f[48 + eax * 4]
				}

			}
		
			_asm {
				
				;clean up terms not in sum
				addps	xmm0, f[0x2708 * 4]
				addps	xmm5, f[16 + 0x2708 * 4]
				
				;accumulate into mm0
				;and restore eax
				addps   xmm0, xmm4
				addps   xmm6, xmm5
				pop		eax
				addps   xmm0, xmm6
		
			;compute products and complete sums of S0, S1, S2
		
				movaps  xmm2, f[0]
				movaps	xmm4, factvec[0]
				mulps   xmm0, xmm4
				mulps   xmm2, xmm7 
				addps   xmm0, xmm2
				mulps   xmm0, xmm3
				
				;accumulate xmm0 to a scalar result
				movaps  xmm1, xmm0
				shufps  xmm1, xmm0,0xee
				addps   xmm1, xmm0
				movaps  xmm0, xmm1
				shufps  xmm1, xmm1,0xe5
				addps   xmm1, xmm0
				movss   Integral, xmm1
		
			}

		}
		

		//stop the timer
		stoptime = GetTickCount();

		//compute the elapsed time, in seconds
		ElapsedTime = (stoptime - starttime) * .001;

		//display the result
		printf("%s%g%s", "\nElapsed Time = ", ElapsedTime, " seconds\n");
		printf("%s%g%s","Integral = ",Integral,"\n\n");
	}
	
	else{
	
		printf("SSE instructions not available on this CPU\n\n");
		
	}
	
	
}//end fnC()

//entry function
int main()
{
	//local declarations
	//integers
	int bTERM;		//menu flag

	//character strings
	char inchar;		//input character
	

	//initialize bTERM to NOT TERMINATE
	bTERM = 0;
	
	//clear terminal
	system("cls");
	
	do {
		
		//present menu to user
		printf("\nEnter the Number of the routine to run, and press <Enter> \n");
		printf("1\tSimpson's Rule Integration using x87 instructions\n");
		printf("2\tSimpson's Rule Integration using 3dNow! SIMD Instructions\n");
		printf("3\tSimpson's Rule Integration using SSE SIMD Instructions\n");
		printf("q\tEnd this program\n");
	
		//get user input
		scanf("%s", &inchar);

		//call the user selected function
		switch	(inchar){

			case '1': fnA();
				break;
	
			case '2': fnB();
				break;

			case '3': fnC();
				break;
		
			//select the terminate function	
			case 'q': printf("Ending This Program\n");
				
				//set flag to terminate
				bTERM = 1;
				break;

		}
		
	} while(!bTERM);

	return 0;

} // End Main

