///////////////////// genMatrix Implementation //////////////////////
//                                                                 //
//                   x87 based Matrix Routines                     //
//                                                                 //
//  Written by:  John S. Riley                                     //
//                                                                 //
//  This file contains a variety of matrix manipulation routines   //
//  written for generic x87 floating-point processors.             //
//                                                                 //
//  The routines use no explicit SIMD instructions, but may be     //
//  compiled using compiler generated SSE or 3dNow instructions.   //
//                                                                 //
//  The matrix routines are general, using rolled up, variable     //
//  sized loops.  Matrix dimension and data are passed to the      //
//  functions as parameters (data or pointers).                    //
//                                                                 //
/////////////////////////////////////////////////////////////////////

///////////////////////// Functions in this File ////////////////////
//                                                                 //
//  genMatrixMult:      Matrix Multiplication                      //
//	genMatrixMultnxn:	Matrix Multiplication specific for nxn     //
//  genMatrixNorm:      Find the Matrix Norm                       //
//  genMatrixTrace:     Calculate the Trace of a nxn Matrix        //
//  genMatrixTranspose: Transpose a matrix                         //
//  genMatrixScale:     Scale a matrix; multiply by a scalar       //
//  genMatrixVectorMult:Multiply a matrix row by a column vector   //
//  genVectorMatrixMult:Multiply a row vector by a matrix column   //
//  genMatrixRotate:    Compute matrix after rotation of axes      //
//  genMatrixTranslate: Compute matrix after translaton of axes    //
//                                                                 //
/////////////////////////////////////////////////////////////////////

///////////////////////// INCLUDES //////////////////////////////////

#include "genMatrix.h"
#include "stdlib.h"
#include "math.h"

///////////////////////// FUNCTIONS /////////////////////////////////

//////////////////////// genMatrixScale /////////////////////////////
//																   //
//	This function multiplies a single precision mxn matrix		   //
//  by a scalar.												   //
//																   //
//	R = sA computed as Rmn = s * Amn							   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*MatrixA	float	Pointer to A data				   //		   //
//			DimenM		int		m dimension of A; rows			   //
//			DimenN		int		n dimension of A; columns		   //
//			Scale		float	scalar factor, s				   //
//			*Result		float	Pointer to Matrix R = sA		   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genMatrixScale(
				   float *MatrixA, 
				   int DimenM, 
				   int DimenN, 
				   float Scale,
				   float *Result)
{

	//local declarations
	//integers
	int n;		//counter for loop over columns
	int m;		//counter for loop over rows
	int indx;	//pointer offset from A[00]

	//loop over columns
	for(m=0; m < DimenM; ++m){

		//loop over rows
		for(n=0; n < DimenN; ++n){

			//compute pointer offset
			indx = m * DimenM + n;

			//scale matrix element
			*(Result + indx) = *(MatrixA + indx) * Scale;
		
		}
	}

	//return any math error
	return(errno);

} // End genMatrixScale

//////////////////////// genMatrixMultnxn ///////////////////////////
//																   //
//	This function multiplies two nxn matrices, A and B			   //												   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*MatrixA	float	Pointer to A data				   //		   //
//			*MatrixB	float	Pointer to B data    			   //
//			dimen		int		n dimension of A, B				   //
//			*MatrixC	float	Pointer to Matrix C = A x B		   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genMatrixMultnxn(
					 float *MatrixA,
					 float *MatrixB,
					 int	dimen,
					 float *MatrixC)

{

	//local declarations
	//integers
	int i;			//counter for rows for matrix C
	int j;			//counter for over columns of matrix C
	int k;			//counter for row,column of matrices A,B
	int indxij;		//pointer offset for ij-th element
	int indxik;		//pointer offset for ik-th element
	int indxkj;		//pointer offset for kj-th element

	//loop over rows for matrix C
	for(i=0; i < dimen; ++i){

		//loop over columns of matrix C
		for(j=0; j < dimen; ++j){

			//compute offset of ij-th element
			indxij = i * dimen + j;

			//zero ij-th element of result matrix, C
			*(MatrixC + indxij) = 0;

			//loop over row,column of matrices A,B
			for(k=0; k < dimen; ++k){

				//compute offset for ik-th element for matrix A
				indxik = i * dimen + k;

				//compute offset for kj-th element for matrix B
				indxkj = k * dimen + j;

				//build c[ij] from sum{a[ik] * b[kj]}
				*(MatrixC + indxij) += *(MatrixA + indxik) * *(MatrixB + indxkj);

			}

		}

	}

	//return any math error
	return(errno);

} //End genMatrixMultnxn

//////////////////////// genMatrixTrace  ////////////////////////////
//																   //
//	This function computes the trace of an nxn matrix, A		   //												   //
//																   //
//	The trace is computed as Trace = sum[a(ii)], the sum of the	   //
//	diagonal elements.											   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*Matrix		float	Pointer to matrix data			   //		   
//			dimen		int		n dimension of A				   //
//			*trace		float	Pointer to where trace is stored   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genMatrixTrace(
				   float *Matrix,
				   int	 dimen,
				   float *trace)
{

	//local declarations
	//integers
	int i;			//counter to loop over diagonal elements of A
	int indx;		//pointer offset for elements of matrix A

	//zero the trace
	*trace = 0;

	//loop over diagonal elements
	for(i=0; i < dimen; ++i){

		//compute the offset of the ii-th element of A
		indx = i * dimen + i;

		//build the trace = sum{a[ii]}
		*trace += *(Matrix + indx);

	}

	//return any math error
	return(errno);

} //End genMatrixTrace

//////////////////////// genMatrixMult //////////////////////////////
//																   //
//	This function multiplies an mxn matrix by an nxp matrix		   //												   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*MatrixA	float	Pointer to A data				   //
//			dimM		int		rows in matrix A				   //
//			dimN		int		columns in matrix A				   //
//			*MatrixB	float	Pointer to B data    			   //
//			dimP		int		columns of matrix B(rows = dimN)   //
//			*MatrixC	float	Pointer to Matrix C = A x B		   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genMatrixMult(
					 float *MatrixA,
					 int	dimM,
					 int	dimN,
					 float *MatrixB,
					 int	dimP,
					 float *MatrixProd)

{

	//local declarations
	//integers
	int j;			//row counter for C matrix
	int k;			//column counter for C matrix
	int n;			//row,column counter matrices A,B
	int indxjn;		//pointer offset for jn-th element of matrix A
	int indxnk;		//pointer offset for nk-th element of matrix B
	int indxjk;		//pointer offset for jk-th element of matrix C

	//loop over rows of C
	for(j=0; j < dimM; ++j){

		//loop over columns of C
		for(k=0; k < dimP; ++k){

			//compute offset for jk-th for element of C
			indxjk = j * dimM + k;

			//zero c[jk]
			*(MatrixProd + indxjk) = 0;

			//loop over row,column of A,B
			for(n=0; n < dimN; ++n){

				//compute offset for jn-th element of A
				indxjn = j * dimM + n;

				//compute offset for nk-th element of B
				indxnk = n * dimM + k;

				//build c[jk] = sum{a[jn] * b[nk]
				*(MatrixProd + indxjk) += *(MatrixA + indxjn) * *(MatrixB + indxnk);

			}

		}

	}

	//return any math error
	return(errno);

} //End genMatrixMult

//////////////////////// genMatrixAdd ///////////////////////////////
//																   //
//	This function adds two mxn matrices, A and B, such that		   //
//	the sum, C is given by c[ij] = a[ij] + b[ij]				   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*MatrixA	float	Pointer to A data				   //		   //
//			*MatrixB	float	Pointer to B data    			   //
//			dimM		int		m dimension of A, B; rows		   //
//			dimN		int		n dimension of A, B; columns	   //
//			*MatrixSum	float	Pointer to store MatrixSum= A + B  //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int	genMatrixAdd(
				 float	*MatrixA,
				 float	*MatrixB,
				 int	dimM,
				 int	dimN,
				 float	*MatrixSum)
{

	//local declarations
	//integers
	int m;			//counter to loop over rows
	int n;			//counter to loop over columns
	int indx;		//pointer offset to index mn-th elements

	//loop over rows
	for(m=0; m < dimM; ++m){

		//loop over columns
		for(n=0; n < dimN; ++n){

			//compute offset for mn-th elements
			indx = m * dimM + n;
			
			//sum[mn] = a[mn] + b[mn] 
			*(MatrixSum + indx) = *(MatrixA + indx) + *(MatrixB + indx);

		}

	}

	//return any math error
	return(errno);

}  // End genMatrixAdd

//////////////////////// genMatrixTranspose /////////////////////////
//																   //
//	This computes the transpose of an mxn matrix A, where		   //
//	at[nm] = a[mn]												   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*Matrix		float	Pointer to A data				   //
//			dimM		int		m dimension of A, B; rows		   //
//			dimN		int		n dimension of A, B; columns	   //
//			*Transpose	float	Pointer to store Transpose of A    //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genMatrixTranspose(
					   float	*Matrix,
					   int		dimM,
					   int		dimN,
					   float	*Transpose)
{

	//local declarations
	//integers
	int j;				//row counter
	int k;				//column counter

	int RowSel;			//holder for row part of offset
	int indxjk;			//pointer offset for jk-th element
	int indxkj;			//pointer offset for kj-th element

	//loop over rows
	for(j=0; j < dimM; ++j){

		//compute the row part of offset for Matrix
		//doing this prevents needless recomputing this 
		//value inside k loop
		RowSel = j * dimM;

		//loop over columns
		for(k=0; k < dimN; ++k){

			//compute offset to jk-th element of matrix
			indxjk = RowSel + k;

			//compute offset to kj-th element for transpose
			indxkj = k * dimN + j;

			//transpose[kj] = matrix[jk]
			*(Transpose + indxkj) = *(Matrix + indxjk);

		}

	}

	//return any math error
	return(errno);

} // End genMatrixTranspose