//////////////////// genSolvers Implementation //////////////////////
//                                                                 //
//                   x87 based Solver Routines                     //
//                                                                 //
//  Written by:  John S. Riley                                     //
//                                                                 //
//  This file contains a variety of solver routines				   //
//  written for generic x87 floating-point processors.             //
//                                                                 //
//  The routines use no explicit SIMD instructions, but may be     //
//  compiled using compiler generated SSE or 3dNow instructions.   //
//                                                                 //
//  The solver routines are general, using rolled up, variable     //
//  sized loops.  Dimension and data are passed to the		       //
//  functions as parameters (data or pointers).                    //
//                                                                 //
/////////////////////////////////////////////////////////////////////

///////////////////////// Functions in this File ////////////////////
//                                                                 //
//  genGaussSeidel:     generic Gauss-Seidel linear system solver  //
//	genRungeKutta:		generic Runge-Kutta solver for y'=f(x,y)   //
//                                                                 //
/////////////////////////////////////////////////////////////////////

//////////////////////// INCLUDES //////////////////////////////////

#include "genSolvers.h"
#include "stdlib.h"
#include "math.h"

///////////////////////// DEFINES ///////////////////////////////////

#define	one_sixth	0.16666667

///////////////////////// FUNCTIONS /////////////////////////////////

//////////////////////// genGaussSeidel /////////////////////////////
//																   //
//	This function solves the matrix-vector equation At = b for     //
//	the vector t using a generic Gauss-Seidel iteration.		   //
//																   //
//	The function will perform a maximum of 1000 iterations.		   //
//																   //
//	INPUT: Passed as parameters:								   //
//																   //
//			*a			float	Pointer to a matrix data		   //
//			*t			float	Pointer to t vector data		   //
//			*b			float	Pointer to b vector data		   //
//			n			int		dimension vectors and matrix	   //
//			*t_soln		float	Pointer to vector to hold solution //
//			OptTol		float	tolerance for convergence		   //
//																   //
//	OUTPUT:	Return value is NumIts, the number of iterations	   //
//			executed before exit.  User should test the return     //
//			value to ensure the system actually converged.		   //
//																   //
/////////////////////////////////////////////////////////////////////

int genGaussSeidel(
				   float	*a,
				   float	*t,
				   float	*b,
				   int		n,
				   float	OptTol)
{
	//local declarations
	//integers
	int j;					//row counter in Gauss-Seidel solver
	int k;					//column counter in Gauss-Seidel solver
	int *indx;				//pointer math offset for matrix row
	int NumIts;				//number of iterations in Gauss-Seidel solver
							//this is used to prevent infinite loop in case
							//of poor or non-convergence in the solver

	//floats
	float delta_T;			//max change in Temperature; used to test for convergence
	float max_delta_T;		//used to determine the delta_T
	float sum1;				//a temporary summation holder
	float sum2;				//a temporary summation holder
	float Tdiff;			//individual Temperature difference; used
							//to compute max_delta_T
	float *t_old;			//pointer to temporary 'last' iteration
	float *ajj_inv;			//pointer to -1/a[j][j]


	//allocate for t_old
	t_old = malloc(n * 4);

	//allocate for -1/a[j][j]
	ajj_inv = malloc(n * 4);

	//allocate for indx[j]
	indx = malloc(n * 4);

	//solve the system for t(j) using Gauss-Seidel procedure

	//initialize values need for determining convergence
	NumIts = 1;
	delta_T = 1;
	max_delta_T = 0;

	
	
	for(j=0; j < n; ++j){

		//save guess values of t(j)'s into current t(j)'s
		*(t_old + j) = *(t + j);

		//calculate row pointer once outside iteration
		*(indx + j) = j * n;

		//calculate -1/a[j][j] once outside iteration
		*(ajj_inv + j) = -1 / (*(a + *(indx + j) + j));

	}


	do{

		for(j=0; j < n; ++j){
			
			//update t(j) part using t(i)'s (i<j) already updated

			//zero sum1 since running sum
			sum1 = 0;

			//compute offset for row
			//indx = j * n;

			for(k=0; k <= j-1; ++k){

				sum1 += *(a + *(indx + j) + k) * *(t + k);

			}

			//update t(j) part using t(i)'s (i>j) not yet updated
			//this iteration

			//zero sum2 since running sum
			sum2 = 0;

			for(k = j+1; k < n; ++k){

				sum2 += *(a + *(indx + j) + k) * *(t_old + k);

			}

			*(t + j) = sum1 + sum2 - *(b + j);
			*(t + j) *= *(ajj_inv + j);

			//find max change for this set of t(j)
			Tdiff = (float)fabs(*(t + j) - *(t_old + j));

			if(Tdiff > max_delta_T){

				max_delta_T = Tdiff;

			}

		}

		//increment number of iterations
		++NumIts;

		//reset t change values used for convergence testing
		delta_T = max_delta_T;
		max_delta_T = 0;

		//save the old t(j)'s
		for(j = 0; j < n; ++j){

			*(t_old + j) = *(t + j);

		}

	}while ((delta_T >= OptTol) & (NumIts <= 1000));

	//free memory allocated for local arrays
	free(t_old);
	free(ajj_inv);
	free(indx);

	return(NumIts);

} //end genGaussSeidel

//////////////////////// genRungeKutta //////////////////////////////
//																   //
//	This function solves first order differential equations of the //
//	for y'=f(x,y) for the function y using the Runge-Kutta		   //
//	algorithm, given the initial conditions x0 and y0.    		   //
//																   //
//	The user must include code in the calling program for the	   //
//	function f(x,y) with the function and both arguments declared  //
//	as doubles.  A pointer to this function is passed to the	   //
//	genRungeKutta routine.										   //
//																   // 
//	INPUT: Passed as parameters:								   //
//																   //
//			x			double	initial x value					   //
//			*y			double	Pointer to y data (see note below) //
//			(*fx)		double	Function Pointer to f(x,y)		   //
//			h			double	x value step sixe				   //
//			n			int		number of x,y values to solve	   //
//																   //
//		NOTE:	The first entry of the array y should be y0 as	   //
//				specified by the initial conditions of the		   //
//				problem.  The remainder of this array will be	   //
//				overwritten.  The user should initialize this	   //
//				array to be large enough, specifically as		   //
//																   //
//				float y[n+1];  in the calling program.			   //
//																   //
//	OUTPUT:	Return value is errno, math error code defined in	   //
//			C standard library.									   //
//																   //
//			Returns 0 if no math error.							   //
//																   //
/////////////////////////////////////////////////////////////////////

int genRungeKutta(
				  double	x,
				  double	*y,
				  double	(*fx)(double, double),
				  double	h,
				  int	n)
{

	//local declarations
	//integers
	int	j;					//function value counter

	//doubles
	double	k1;				//Runge Kutta k1 term
	double	k2;				//Runge Kutta k2 term
	double	k3;				//Runge Kutta k3 term
	double	k4;				//Runge Kutta k4 term
	double	fn;				//return value for call to f(x,y)
	double	h_2;			//h/2
	double	yn;				//'current' y(x)
	double	xarg;			//x value sent to f(x,y)
	double	yarg;			//y value sent to f(x,y)

	//compute h/2
	h_2 = (0.5 * h);

	//solve for each value of x
	for(j=0; j < n; ++j){

		//cache 'current' yn
		yn = *(y + j);

		//get f(x,y) for k1
		fn = fx(x,yn);
		k1 = h * fn;

		//compute x and y for k2
		xarg = (x + h_2);
		yarg = (yn + (0.5 * k1));

		//get f(x,y) for k2
		fn = fx(xarg, yarg);
		k2 = h * fn;

		//compute y for k3; xarg is same as for k2
		yarg = yn + (0.5 * k2);

		//get f(x,y) for k3
		fn = fx(xarg, yarg);
		k3 = h * fn;

		//compute x and y for k4
		xarg = x + h;
		yarg = yn + k3;

		//get f(x,y) for k4
		fn = fx(xarg, yarg);
		k4 = h * fn;

		//compute y(n+1) = yn + 1/6 * (k1+ 2k2 + 2k3 + k4)
		k2 *= 2;
		k3 *= 2;
		k2 += k3;
		k1 += k2 + k4;

		*(y+j+1) = yn + one_sixth * k1;

		//increment x by the step size
		x += h;

	}

	return(errno);

} // End genRungeKutta

